import json
import os
from distutils.dir_util import copy_tree
from pathlib import Path
from types import SimpleNamespace
from unittest import mock
from unittest.mock import patch

import bigeye_sdk
import pytest
from bigeye_sdk.generated.com.bigeye.models.generated import MetricRunStatus
from click.testing import CliRunner

from bigquery_etl.cli.monitoring import (
    delete,
    deploy,
    deploy_custom_rules,
    run,
    set_partition_column,
    update,
    validate,
)

TEST_DIR = Path(__file__).parent.parent


class TestMonitoring:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    @pytest.fixture(autouse=True)
    def set_api_key(self):
        os.environ["BIGEYE_API_KEY"] = "test-api-key"

    @patch("bigeye_sdk.client.datawatch_client.datawatch_client_factory")
    @patch(
        "bigeye_sdk.controller.metric_suite_controller.MetricSuiteController.__init__"
    )
    def test_deploy(self, mock_metric_controller_init, mock_client_factory, runner):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "incremental_query_v1"
            )

            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/incremental_query_v1")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))

            mock_metric_controller_init.return_value = None
            mock_client_factory.return_value = None
            with patch.object(
                bigeye_sdk.controller.metric_suite_controller.MetricSuiteController,
                "execute_bigconfig",
            ) as mock_execute_bigconfig:
                mock_execute_bigconfig.return_value = None
                runner.invoke(deploy, [f"{str(SQL_DIR)}"])

                mock_execute_bigconfig.assert_called_once()
                assert (SQL_DIR / "bigconfig.yml").exists()

    @patch("bigquery_etl.cli.monitoring.datawatch_client_factory")
    @patch(
        "bigeye_sdk.controller.metric_suite_controller.MetricSuiteController.__init__"
    )
    def test_deploy_custom_rule(
        self, mock_metric_controller_init, mock_client_factory, runner
    ):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "incremental_query_v1"
            )

            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/incremental_query_v1")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))

            (SQL_DIR / "bigeye_custom_rules.sql").write_text(
                """
                -- {
                --   "name": "Custom check",
                --   "alert_conditions": "value",
                --   "range": {
                --     "min": 0,
                --     "max": 1
                --   },
                --   "collections": ["Test"],
                --   "owner": "",
                --   "schedule": "Default Schedule - 13:00 UTC"
                -- }
                SELECT
                COUNT(*)
                FROM
                `{{ project_id }}.{{ dataset_id }}.{{ table_name }}`;
            """
            )

            mock_metric_controller_init.return_value = None
            mock_client = mock.Mock()
            mock_client_factory.return_value = mock_client
            mock_client.get_collections.return_value = mock.Mock(collections=[])
            mock_client.get_rules_for_source.return_value = mock.Mock(custom_rules=[])
            mock_client.get_named_schedule.return_value = mock.Mock(named_schedules=[])
            mock_client._call_datawatch.return_value = None

            runner.invoke(deploy_custom_rules, [f"{str(SQL_DIR)}"])

            # mock_client_factory._call_datawatch.assert_called_once()
            mock_client.get_collections.assert_called_once()
            mock_client.get_rules_for_source.assert_called_once()
            mock_client._call_datawatch.assert_called_once()
            expected_body = {
                "customRule": {
                    "name": "Custom check",
                    "sql": '/* { */ /*   "name": "Custom check", */ /*   "alert_conditions": "value", */ /*   '
                    + '"range": { */ /*     "min": 0, */ /*     "max": 1 */ /*   }, */ /*   "collections": ["Test"], */ /*   '
                    + '"owner": "", */ /*   "schedule": "Default Schedule - 13:00 UTC" */ /* } '
                    + "*/ SELECT COUNT(*) FROM `moz-fx-data-shared-prod.test.incremental_query_v1`",
                    "warehouseId": 1939,
                    "thresholdType": "CUSTOM_RULES_THRESHOLD_TYPE_COUNT",
                    "lowerThreshold": 0,
                    "upperThreshold": 1,
                    "owner": {"email": ""},
                    "collectionIds": [],
                }
            }
            assert mock_client._call_datawatch.call_args.kwargs["body"] == json.dumps(
                expected_body
            )

    def test_update(self, runner):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "incremental_query_v1"
            )

            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/incremental_query_v1")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))

            assert not (SQL_DIR / "bigconfig.yml").exists()
            runner.invoke(update, [f"{str(SQL_DIR)}"])
            assert (SQL_DIR / "bigconfig.yml").exists()

    def test_update_existing_bigconfig(self, runner):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "incremental_query_v1"
            )

            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/incremental_query_v1")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))
            (SQL_DIR / "bigconfig.yml").write_text(
                """
                type: BIGCONFIG_FILE
                table_deployments:
                - deployments:
                - fq_table_name: moz-fx-data-shared-prod.moz-fx-data-shared-prod.test.incremental_query_v1
                    table_metrics:
                    - metric_type:
                        type: PREDEFINED
                        predefined_metric: FRESHNESS
                    metric_schedule:
                        named_schedule:
                        name: Default Schedule - 13:00 UTC
            """
            )

            assert (SQL_DIR / "bigconfig.yml").exists()
            runner.invoke(update, [f"{str(SQL_DIR)}"])
            assert (SQL_DIR / "bigconfig.yml").exists()
            print((SQL_DIR / "bigconfig.yml").read_text())
            assert (
                "predefined_metric: FRESHNESS"
                in (SQL_DIR / "bigconfig.yml").read_text()
            )
            assert (
                "predefined_metric: VOLUME"
                not in (SQL_DIR / "bigconfig.yml").read_text()
            )

    def test_validate_no_bigconfig_file(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/test_no_file")
            os.makedirs(str(SQL_DIR))

            assert not (SQL_DIR / "bigconfig.yml").exists()
            result = runner.invoke(validate, [f"{str(SQL_DIR)}"])
            assert result.exit_code == 0

    def test_validate_empty_file(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/test_empty")
            os.makedirs(str(SQL_DIR))
            (SQL_DIR / "bigconfig.yml").write_text("")

            result = runner.invoke(validate, [f"{str(SQL_DIR)}"])
            assert result.exit_code == 1

    def test_validate_invalid_file(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/invalid")
            os.makedirs(str(SQL_DIR))
            (SQL_DIR / "bigconfig.yml").write_text("invalid")

            result = runner.invoke(validate, [f"{str(SQL_DIR)}"])
            assert result.exit_code == 1

    @patch("bigquery_etl.cli.monitoring.datawatch_client_factory")
    @patch(
        "bigeye_sdk.controller.metric_suite_controller.MetricSuiteController.__init__"
    )
    def test_set_partition_column(
        self, mock_metric_controller_init, mock_client_factory, runner
    ):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "view_with_metadata"
            )
            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/view_with_metadata")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))

            mock_metric_controller_init.return_value = None
            mock_client = mock.Mock()
            mock_client_factory.return_value = mock_client
            mock_client.get_table_ids.return_value = [1234]
            mock_client.get_columns.return_value = mock.Mock(
                columns=[SimpleNamespace(column=SimpleNamespace(name="date", id=111))]
            )
            call_datawatch_mock = mock.Mock()
            mock_client._call_datawatch.return_value = call_datawatch_mock
            call_datawatch_mock.get.return_value = {
                "datasetId": 1234,
                "requiredPartitionColumnId": 111,
            }

            runner.invoke(
                set_partition_column, [f"{str(SQL_DIR)}"], catch_exceptions=False
            )

            mock_client._call_datawatch.assert_called_once()
            expected_body = {"columnId": 111}
            assert mock_client._call_datawatch.call_args.kwargs["body"] == json.dumps(
                expected_body
            )

    @patch("bigquery_etl.cli.monitoring.datawatch_client_factory")
    @patch(
        "bigeye_sdk.controller.metric_suite_controller.MetricSuiteController.__init__"
    )
    def test_delete(self, mock_metric_controller_init, mock_client_factory, runner):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "incremental_query_v1"
            )

            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/incremental_query_v1")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))

            mock_metric_controller_init.return_value = None
            mock_client = mock.Mock()
            mock_client_factory.return_value = mock_client
            mock_client.get_rules_for_source.return_value = mock.Mock(custom_rules=[])
            mock_client.get_metric_info_batch_post.return_value = mock.Mock(
                metrics=[1234]
            )
            mock_client.delete_metrics.return_value = None

            runner.invoke(
                delete, [f"{str(SQL_DIR)}", "--metrics"], catch_exceptions=False
            )
            mock_client.delete_metrics.assert_called_once_with(metrics=[1234])

    @patch("bigquery_etl.cli.monitoring.datawatch_client_factory")
    @patch(
        "bigeye_sdk.controller.metric_suite_controller.MetricSuiteController.__init__"
    )
    def test_delete_custom_sql(
        self, mock_metric_controller_init, mock_client_factory, runner
    ):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "incremental_query_v1"
            )

            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/incremental_query_v1")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))
            (SQL_DIR / "bigeye_custom_rules.sql").write_text("SELECT 2")

            mock_metric_controller_init.return_value = None
            mock_client = mock.Mock()
            mock_client_factory.return_value = mock_client
            rules_mock = mock.Mock(
                custom_rules=[
                    mock.Mock(custom_rule=SimpleNamespace(id=1, sql="SELECT 2"))
                ]
            )
            mock_client.get_rules_for_source.return_value = rules_mock
            mock_client.delete_custom_rule.return_value = None
            mock_client.delete_metrics.return_value = None

            runner.invoke(
                delete,
                [f"{str(SQL_DIR)}", "--custom-sql"],
                catch_exceptions=False,
            )

            mock_client.delete_metrics.assert_not_called()
            mock_client.delete_custom_rule.assert_called_once_with(
                rules_mock.custom_rules[0].id
            )

    @patch("bigquery_etl.cli.monitoring.datawatch_client_factory")
    @patch(
        "bigeye_sdk.controller.metric_suite_controller.MetricSuiteController.__init__"
    )
    def test_run(self, mock_metric_controller_init, mock_client_factory, runner):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "incremental_query_v1"
            )

            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/incremental_query_v1")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))

            mock_metric_controller_init.return_value = None
            mock_client = mock.Mock()
            mock_client_factory.return_value = mock_client
            mock_client.run_metric_batch_async.return_value = []
            mock_client.get_rules_for_source.return_value = mock.Mock(custom_rules=[])
            mock_metric_info = mock.Mock(
                metrics=[
                    SimpleNamespace(
                        name="test", metric_configuration=SimpleNamespace(id=1234)
                    )
                ]
            )
            mock_client.get_metric_info_batch_post.return_value = mock_metric_info

            runner.invoke(
                run,
                [f"{str(SQL_DIR)}"],
                catch_exceptions=False,
            )

        mock_client.run_metric_batch_async.assert_called_once_with(
            metric_ids=[mock_metric_info.metrics[0].metric_configuration.id]
        )

    @patch("bigquery_etl.cli.monitoring.datawatch_client_factory")
    @patch(
        "bigeye_sdk.controller.metric_suite_controller.MetricSuiteController.__init__"
    )
    def test_run_fail(self, mock_metric_controller_init, mock_client_factory, runner):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "incremental_query_v1"
            )

            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/incremental_query_v1")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))

            mock_metric_controller_init.return_value = None
            mock_client = mock.Mock()
            mock_client_factory.return_value = mock_client
            mock_client.run_metric_batch_async.return_value = [
                SimpleNamespace(
                    latest_metric_runs=[
                        SimpleNamespace(
                            status=MetricRunStatus.METRIC_RUN_STATUS_UPPERBOUND_CRITICAL
                        )
                    ],
                    metric_configuration=SimpleNamespace(id=123, name="test [fail]"),
                    active_issue=SimpleNamespace(display_name="error"),
                )
            ]
            mock_client.get_rules_for_source.return_value = mock.Mock(custom_rules=[])
            mock_metric_info = mock.Mock(
                metrics=[
                    SimpleNamespace(
                        name="test", metric_configuration=SimpleNamespace(id=1234)
                    )
                ]
            )
            mock_client.get_metric_info_batch_post.return_value = mock_metric_info

            result = runner.invoke(
                run,
                [f"{str(SQL_DIR)}"],
                catch_exceptions=False,
            )
            assert result.exit_code == 1

        mock_client.run_metric_batch_async.assert_called_once_with(
            metric_ids=[mock_metric_info.metrics[0].metric_configuration.id]
        )

    @patch("bigquery_etl.cli.monitoring.datawatch_client_factory")
    @patch(
        "bigeye_sdk.controller.metric_suite_controller.MetricSuiteController.__init__"
    )
    def test_run_warn(self, mock_metric_controller_init, mock_client_factory, runner):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "incremental_query_v1"
            )

            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/incremental_query_v1")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))

            mock_metric_controller_init.return_value = None
            mock_client = mock.Mock()
            mock_client_factory.return_value = mock_client
            mock_client.run_metric_batch_async.return_value = [
                SimpleNamespace(
                    latest_metric_runs=[
                        SimpleNamespace(
                            status=MetricRunStatus.METRIC_RUN_STATUS_UPPERBOUND_CRITICAL
                        )
                    ],
                    metric_configuration=SimpleNamespace(id=123, name="test [warn]"),
                    active_issue=SimpleNamespace(display_name="error"),
                )
            ]
            mock_client.get_rules_for_source.return_value = mock.Mock(custom_rules=[])
            mock_metric_info = mock.Mock(
                metrics=[
                    SimpleNamespace(
                        name="test [warn]",
                        metric_configuration=SimpleNamespace(id=1234),
                    )
                ]
            )
            mock_client.get_metric_info_batch_post.return_value = mock_metric_info

            result = runner.invoke(
                run,
                [f"{str(SQL_DIR)}"],
                catch_exceptions=False,
            )
            assert result.exit_code == 0

        mock_client.run_metric_batch_async.assert_called_once_with(
            metric_ids=[mock_metric_info.metrics[0].metric_configuration.id]
        )

    @patch("bigquery_etl.cli.monitoring.datawatch_client_factory")
    @patch(
        "bigeye_sdk.controller.metric_suite_controller.MetricSuiteController.__init__"
    )
    def test_run_custom_rule(
        self, mock_metric_controller_init, mock_client_factory, runner
    ):
        with runner.isolated_filesystem():
            test_query = (
                TEST_DIR
                / "data"
                / "test_sql"
                / "moz-fx-data-test-project"
                / "test"
                / "incremental_query_v1"
            )

            SQL_DIR = Path("sql/moz-fx-data-shared-prod/test/incremental_query_v1")
            os.makedirs(str(SQL_DIR))
            copy_tree(str(test_query), str(SQL_DIR))
            (SQL_DIR / "bigeye_custom_rules.sql").write_text(
                """
                SELECT 1
            """
            )

            mock_metric_controller_init.return_value = None
            mock_client = mock.Mock()
            mock_client_factory.return_value = mock_client
            mock_client.run_metric_batch_async.return_value = []
            rules_mock = mock.Mock(
                custom_rules=[
                    mock.Mock(
                        custom_rule=SimpleNamespace(id=1, sql="SELECT 1", name="rule")
                    )
                ]
            )
            mock_client.get_rules_for_source.return_value = rules_mock
            mock_metric_info = mock.Mock(metrics=[])
            mock_client.get_metric_info_batch_post.return_value = mock_metric_info
            mock_datawatch = mock.Mock()
            mock_client._call_datawatch.return_value = mock_datawatch
            mock_datawatch.get.return_value = []

            runner.invoke(
                run,
                [f"{str(SQL_DIR)}"],
                catch_exceptions=False,
            )

            mock_client._call_datawatch.assert_called_once()
            assert (
                str(rules_mock.custom_rules[0].id)
                in mock_client._call_datawatch.call_args.kwargs["url"]
            )


class MockBigeyeClient:
    def __init__(*args, **kwargs):
        pass

    def collections(self):
        return



#!/usr/bin/env python3
"""Copy a range of ga_sessions_* tables from one project to another."""
import datetime

import click
from google.cloud import bigquery, exceptions


def get_qualified_table_name(project, ga_id, table_date):
    """Get qualified name for ga_sessions table."""
    return (
        f"{project}.{ga_id}.ga_sessions_"
        f"{datetime.datetime.strftime(table_date, '%Y%m%d')}"
    )


def copy_single_table(bq_client, src_table, dst_table, overwrite):
    """Copy a single day of ga_sessions."""
    job_config = bigquery.CopyJobConfig(
        write_disposition=(
            bigquery.WriteDisposition.WRITE_TRUNCATE
            if overwrite
            else bigquery.WriteDisposition.WRITE_EMPTY
        )
    )
    try:
        copy_job = bq_client.copy_table(
            sources=src_table,
            destination=dst_table,
            job_config=job_config,
        )
    except exceptions.NotFound:
        print(f"{src_table} not found, copy skipped")
        return

    try:
        copy_job.result()
    except exceptions.Conflict:
        print(f"{dst_table} already exists, copy skipped")


@click.command()
@click.option("--start-date", required=True, type=datetime.datetime.fromisoformat)
@click.option("--end-date", type=datetime.datetime.fromisoformat)
@click.option("--src-project", required=True)
@click.option("--dst-project", required=True)
@click.option("--overwrite/--no-overwrite", default=False)
@click.argument("ga-ids", nargs=-1)
def copy_ga_sessions(ga_ids, start_date, end_date, src_project, dst_project, overwrite):
    """Copy ga_sessions tables for one or multiple properties over a date range."""
    bq_client = bigquery.Client(project=dst_project)

    if end_date is None:
        end_date = start_date

    for ga_id in ga_ids:
        bq_client.create_dataset(f"{dst_project}.{ga_id}", exists_ok=True)

        for table_date in [
            end_date - datetime.timedelta(days=i)
            for i in range((end_date - start_date).days + 1)
        ]:
            src_table = get_qualified_table_name(src_project, ga_id, table_date)
            dst_table = get_qualified_table_name(dst_project, ga_id, table_date)

            print(f"copying {src_table} to {dst_table}")
            copy_single_table(bq_client, src_table, dst_table, overwrite)


if __name__ == "__main__":
    copy_ga_sessions()



"""GLEAN Usage."""

from functools import cache, partial
from pathlib import Path

import click
from pathos.multiprocessing import ProcessingPool

from bigquery_etl.cli.utils import (
    is_valid_project,
    table_matches_patterns,
    use_cloud_function_option,
)
from bigquery_etl.config import ConfigLoader
from bigquery_etl.dryrun import get_id_token
from sql_generators.glean_usage import (
    baseline_clients_daily,
    baseline_clients_first_seen,
    baseline_clients_last_seen,
    clients_last_seen_joined,
    event_error_monitoring,
    event_flow_monitoring,
    event_monitoring_live,
    events_stream,
    events_unnested,
    glean_app_ping_views,
    metrics_clients_daily,
    metrics_clients_last_seen,
)
from sql_generators.glean_usage.common import get_app_info, list_tables

# list of methods for generating queries
GLEAN_TABLES = [
    glean_app_ping_views.GleanAppPingViews(),
    baseline_clients_daily.BaselineClientsDailyTable(),
    baseline_clients_first_seen.BaselineClientsFirstSeenTable(),
    baseline_clients_last_seen.BaselineClientsLastSeenTable(),
    events_unnested.EventsUnnestedTable(),
    metrics_clients_daily.MetricsClientsDaily(),
    metrics_clients_last_seen.MetricsClientsLastSeen(),
    clients_last_seen_joined.ClientsLastSeenJoined(),
    event_monitoring_live.EventMonitoringLive(),
    event_error_monitoring.EventErrorMonitoring(),
    event_flow_monitoring.EventFlowMonitoring(),
    events_stream.EventsStreamTable(),
]


@click.command()
@click.option(
    "--target-project",
    "--target_project",
    help="GCP project ID",
    default="moz-fx-data-shared-prod",
    callback=is_valid_project,
)
@click.option(
    "--output-dir",
    "--output_dir",
    help="Output directory generated SQL is written to",
    type=click.Path(file_okay=False),
    default="sql",
)
@click.option(
    "--parallelism",
    "-p",
    help="Maximum number of tasks to execute concurrently",
    default=8,
)
@click.option(
    "--except",
    "-x",
    "exclude",
    help="Process all tables except for the given tables",
)
@click.option(
    "--only",
    "-o",
    help="Process only the given tables",
)
@click.option(
    "--app_name",
    "--app-name",
    help="App to generate per-app dataset metadata and union views for.",
)
@use_cloud_function_option
def generate(
    target_project, output_dir, parallelism, exclude, only, app_name, use_cloud_function
):
    """Generate per-app_id queries and views, and per-app dataset metadata and union views.

    Note that a file won't be generated if a corresponding file is already present
    in the target directory, which allows manual overrides of generated files by
    checking them into the sql/ tree of the default branch of the repository.
    """
    table_filter = partial(table_matches_patterns, "*", False)

    if only:
        table_filter = partial(table_matches_patterns, only, False)
    elif exclude:
        table_filter = partial(table_matches_patterns, exclude, True)

    @cache
    def get_tables(table_name="baseline_v1"):
        baseline_tables = list_tables(
            project_id=target_project,
            only_tables=[only] if only else None,
            table_filter=table_filter,
            table_name=table_name,
        )

        # filter out skipped apps
        return [
            baseline_table
            for baseline_table in baseline_tables
            if baseline_table.split(".")[1]
            not in [
                f"{skipped_app}_stable"
                for skipped_app in ConfigLoader.get(
                    "generate", "glean_usage", "skip_apps", fallback=[]
                )
            ]
        ]

    output_dir = Path(output_dir) / target_project

    # per app specific datasets
    app_info = get_app_info()
    if app_name:
        app_info = {name: info for name, info in app_info.items() if name == app_name}

    app_info = [
        info
        for name, info in app_info.items()
        if name
        not in ConfigLoader.get("generate", "glean_usage", "skip_apps", fallback=[])
    ]

    id_token = get_id_token()

    # Prepare parameters so that generation of all Glean datasets can be done in parallel

    # Parameters to generate per-app_id datasets consist of the function to be called
    # and baseline tables
    generate_per_app_id = [
        (
            partial(
                table.generate_per_app_id,
                target_project,
                output_dir=output_dir,
                use_cloud_function=use_cloud_function,
                app_info=app_info,
                parallelism=parallelism,
                id_token=id_token,
            ),
            baseline_table,
        )
        for table in GLEAN_TABLES
        for baseline_table in get_tables(table_name=table.base_table_name)
    ]

    # Parameters to generate per-app datasets consist of the function to be called
    # and app_info
    generate_per_app = [
        (
            partial(
                table.generate_per_app,
                target_project,
                output_dir=output_dir,
                use_cloud_function=use_cloud_function,
                parallelism=parallelism,
                id_token=id_token,
            ),
            info,
        )
        for info in app_info
        for table in GLEAN_TABLES
    ]

    # Parameters to generate datasets that union all app datasets
    generate_across_apps = [
        (
            partial(
                table.generate_across_apps,
                target_project,
                output_dir=output_dir,
                use_cloud_function=use_cloud_function,
                parallelism=parallelism,
            ),
            app_info,
        )
        for table in GLEAN_TABLES
    ]

    with ProcessingPool(parallelism) as pool:
        pool.map(
            lambda f: f[0](f[1]),
            generate_per_app_id + generate_per_app + generate_across_apps,
        )



#!/usr/bin/env python3 -W ignore

"""Generate view to unnest parquet-format list and map fields."""

import json
import sys
from argparse import ArgumentParser
from textwrap import dedent

from google.cloud import bigquery


def qualify(table, dataset, project):
    """Fully qualify table in standard sql format without `s."""
    table = table.replace("`", "").replace(":", ".")
    dots = table.count(".")
    if dots == 0:
        return f"{project}.{dataset}.{table}"
    elif dots == 1:
        return f"{project}.{table}"
    else:
        return table


def replace(field, unnest_layer=0, *prefix):
    """Convert BigQuery field to a SQL expression for use in a REPLACE block."""
    if field.field_type == "RECORD":
        is_list = (
            len(field.fields) == 1
            and field.fields[0].name == "list"
            and field.fields[0].mode == "REPEATED"
            and len(field.fields[0].fields) == 1
            and field.fields[0].fields[0].name == "element"
        )
        is_map = (
            len(field.fields) == 1
            and field.fields[0].name == "key_value"
            and field.fields[0].mode == "REPEATED"
        )
        if is_list or is_map:
            # prevent naming collisions via `UNNEST(...) AS`
            unnest_as = f"_{unnest_layer}"
            if is_list:
                unnest = ".".join(prefix + (field.name, "list"))
                nested = replace(field.fields[0].fields[0], unnest_layer + 1, unnest_as)
                # handle simplest case without unnest_as
                if nested == f"{unnest_as}.element":
                    return f"ARRAY(SELECT * FROM UNNEST({unnest})) AS {field.name}"
            elif is_map:
                unnest = ".".join(prefix + (field.name, "key_value"))
                nested = replace(
                    bigquery.SchemaField(
                        field_type="RECORD",
                        fields=field.fields[0].fields,
                        name=unnest_as,
                    ),
                    unnest_layer + 1,
                )
                # handle simplest case without array unnest_layer
                if nested == unnest_as:
                    return f"{unnest} AS {field.name}"
            # shorten special case for nested struct
            special_prefix = f"(SELECT AS STRUCT {unnest_as}.* "
            if nested.startswith(special_prefix):
                index = len(special_prefix)
                suffix = nested[index:].rsplit(")", 1)[0]
                select = f"SELECT AS STRUCT * {suffix}"
            # shorten nested struct
            elif nested.startswith("(SELECT AS STRUCT"):
                select = nested[1:].rsplit(")", 1)[0]
            # handle nested array
            elif nested.startswith("ARRAY"):
                select = f"SELECT AS STRUCT {nested}"
            else:
                select = f"SELECT {nested}"
            from_ = f"FROM UNNEST({unnest}) AS {unnest_as}"
            return f"ARRAY({select} {from_}) AS {field.name}"
        # handle other structs with nested replacements
        replacements = []
        for subfield in field.fields:
            replacement = replace(subfield, unnest_layer, *prefix, field.name)
            if " AS " in replacement:
                replacements += [replacement]
        if replacements:
            select = f"SELECT AS STRUCT {'.'.join(prefix + (field.name, '*'))}"
            replacements = ", ".join(replacements)
            return f"({select} REPLACE ({replacements})) AS {field.name}"
    # no unnesting needed
    return ".".join(prefix + (field.name,))


def generate(table, view):
    """Generate a CREATE VIEW statement that unnests lists and maps."""
    schema = bigquery.Client().get_table(table).schema
    replacements = []
    for index, field in enumerate(schema):
        try:
            replacement = replace(field)
        except Exception:
            json_field = json.dumps(field.to_api_repr(), indent=2)
            print(f"problem with field {index}:\n{json_field}", file=sys.stderr)
            raise
        if " AS " in replacement:
            replacements += [replacement]

    replacements.sort(key=lambda x: x.rsplit(" AS ", 1)[-1])

    return dedent(
        f"""
        -- Query generated by: {' '.join(sys.argv)}
        CREATE OR REPLACE VIEW
          `{view}` AS
        SELECT
          * REPLACE (
            """
        + f""",
            """.join(
            replacements
        )
        + f"""
          )
        FROM
          `{table}`
        """
    ).strip()


def main():
    """Parse argv and print CREATE VIEW statement."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("view")
    parser.add_argument("table")
    parser.add_argument(
        "--project",
        default="moz-fx-data-shared-prod",
        help="project to use for partially qualified table and view, defaults to"
        " moz-fx-data-shared-prod",
    )
    parser.add_argument(
        "--dataset",
        default="telemetry",
        help="dataset to use for partially qualified table and view, defaults"
        " telemetry",
    )
    args = parser.parse_args()
    view = qualify(args.view, args.dataset, args.project)
    table = qualify(args.table, args.dataset, args.project)
    sql = generate(table, view)
    print(sql)


if __name__ == "__main__":
    main()



"""Glean metric data - download from API, clean and upload to BigQuery."""

import logging
from pathlib import Path

import click
import requests
import yaml
from google.cloud import bigquery

API_BASE_URL = "https://probeinfo.telemetry.mozilla.org"

DEFAULT_PROJECT_ID = Path(__file__).parent.parent.parent.name
DEFAULT_DATASET_ID = Path(__file__).parent.parent.name
DEFAULT_TABLE_NAME = Path(__file__).parent.name
DEFAULT_BAD_REQUEST_THRESHOLD = 5

SCHEMA_FILE = Path(__file__).parent / "schema.yaml"
SCHEMA = bigquery.SchemaField.from_api_repr(
    {"name": "root", "type": "RECORD", **yaml.safe_load(SCHEMA_FILE.read_text())}
).fields


def get_api_response(url):
    """Get json of response if the requests to the API was successful."""
    response = requests.get(url)
    response.raise_for_status()
    return response.json()


def store_data_in_bigquery(data, schema, destination_project, destination_table_id):
    """Upload data to Bigquery in a single, non partitioned table."""
    client = bigquery.Client(project=destination_project)

    job_config = bigquery.LoadJobConfig(
        create_disposition="CREATE_IF_NEEDED",
        schema=schema,
        source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
        write_disposition="WRITE_TRUNCATE",
    )

    load_job = client.load_table_from_json(
        data, destination_table_id, location="US", job_config=job_config
    )
    load_job.result()
    stored_table = client.get_table(destination_table_id)
    logging.info(f"Loaded {stored_table.num_rows} rows into {destination_table_id}.")


def download_glean_metrics(url, threshold):
    """Download metrics for glean products and parse the data."""
    # get a list of all glean apps
    glean_apps_response = get_api_response(f"{url}/glean/repositories")
    glean_apps = [glean_app["name"] for glean_app in glean_apps_response]

    glean_metrics = []
    error_counter = 0
    for glean_app in glean_apps:
        try:
            metrics = get_api_response(f"{url}/glean/{glean_app}/metrics")
            for name, metric in metrics.items():
                first_seen = metric["history"][0]["dates"]["first"][:10]
                last_seen = metric["history"][-1]["dates"]["last"][:10]
                expires = metric["history"][-1]["expires"]
                in_source = metric["in-source"]
                glean_metrics.append(
                    {
                        "glean_app": glean_app,
                        "metric": name,
                        "type": metric["history"][-1]["type"],
                        "first_seen_date": first_seen,
                        "last_seen_date": last_seen,
                        "expires": expires,
                        "in_source": in_source,
                    }
                )
        except requests.exceptions.HTTPError as err:
            error_counter += 1
            logging.error(err)
        if error_counter > threshold:
            raise Exception(
                f"More then the accepted threshold of {threshold} requests failed."
            )
    return glean_metrics


@click.command
@click.option(
    "--bq_project_id",
    default=DEFAULT_PROJECT_ID,
    show_default=True,
    help="BigQuery project the data is written to.",
)
@click.option(
    "--bq_dataset_id",
    default=DEFAULT_DATASET_ID,
    show_default=True,
    help="BigQuery dataset the data is written to.",
)
@click.option(
    "--bq_table_name",
    default=DEFAULT_TABLE_NAME,
    show_default=True,
    help="Bigquery table the data is written to.",
)
@click.option(
    "--threshold",
    default=DEFAULT_BAD_REQUEST_THRESHOLD,
    show_default=True,
    help="Number of bad Requests to get metrics for glean apps before the job fails.",
)
def run_glean_metrics(bq_project_id, bq_dataset_id, bq_table_name, threshold):
    """Download the data from the API and store it in BigQuery."""
    glean_metrics = download_glean_metrics(API_BASE_URL, threshold)

    destination_table_id = f"{bq_project_id}.{bq_dataset_id}.{bq_table_name}"
    store_data_in_bigquery(
        data=glean_metrics,
        schema=SCHEMA,
        destination_project=bq_project_id,
        destination_table_id=destination_table_id,
    )


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.INFO)
    run_glean_metrics()



"""Variables for templated SQL."""

from .utils import compute_datacube_groupings, get_custom_distribution_metadata


def clients_scalar_aggregates(**kwargs):
    """Variables for clients scalar aggregation."""
    attributes_list = [
        "client_id",
        "ping_type",
        "os",
        "app_version",
        "app_build_id",
        "channel",
    ]
    attributes_type_list = ["STRING", "STRING", "STRING", "INT64", "STRING", "STRING"]
    user_data_attributes_list = ["metric", "metric_type", "key"]
    return dict(
        attributes=",".join(attributes_list),
        attributes_list=attributes_list,
        attributes_type=",".join(
            f"{name} {dtype}"
            for name, dtype in zip(attributes_list, attributes_type_list)
        ),
        user_data_attributes=",".join(user_data_attributes_list),
        user_data_type="""
            ARRAY<
                STRUCT<
                metric STRING,
                metric_type STRING,
                key STRING,
                agg_type STRING,
                value FLOAT64
                >
            >
        """,
        **kwargs,
    )


def clients_histogram_aggregates(**kwargs):
    """Variables for histogram aggregates."""
    attributes_list = [
        "sample_id",
        "client_id",
        "ping_type",
        "os",
        "app_version",
        "app_build_id",
        "channel",
    ]
    fixed_attributes = ["app_version", "channel"]
    cubed_attributes = [x for x in attributes_list if x not in fixed_attributes]
    return dict(
        attributes_list=attributes_list,
        attributes=",".join(attributes_list),
        cubed_attributes=cubed_attributes,
        attribute_combinations=compute_datacube_groupings(cubed_attributes),
        metric_attributes="""
            metric,
            metric_type,
            key,
            agg_type
        """,
        **kwargs,
    )


def scalar_bucket_counts(**kwargs):
    """Variables for scalar bucket_counts."""
    attributes_list = ["ping_type", "os", "app_version", "app_build_id", "channel"]
    fixed_attributes = ["app_version", "channel"]
    cubed_attributes = [x for x in attributes_list if x not in fixed_attributes]
    return dict(
        attributes=",".join(attributes_list),
        cubed_attributes=cubed_attributes,
        attribute_combinations=compute_datacube_groupings(cubed_attributes),
        scalar_metric_types="""
            "counter",
            "quantity",
            "labeled_counter",
            "timespan"
        """,
        boolean_metric_types="""
            "boolean"
        """,
        aggregate_attributes="""
            metric,
            metric_type,
            key
        """,
        aggregate_attributes_type="""
            metric STRING,
            metric_type STRING,
            key STRING
        """,
        **{
            # re-use variables from previous query
            key: clients_scalar_aggregates()[key]
            for key in ["user_data_attributes", "user_data_type"]
        },
        **kwargs,
    )


def histogram_bucket_counts(**kwargs):
    """Variables for clients histogram bucket counts."""
    attributes_list = ["ping_type", "os", "app_version", "app_build_id", "channel"]
    metric_attributes_list = ["metric", "metric_type", "key", "agg_type"]
    fixed_attributes = ["app_version", "channel"]
    cubed_attributes = [x for x in attributes_list if x not in fixed_attributes]
    custom_dist_metadata = set()
    for product_name in ["fenix", "firefox-desktop"]:
        custom_dist_metadata.update(get_custom_distribution_metadata(product_name))
    return dict(
        attributes_list=attributes_list,
        attributes=",".join(attributes_list),
        cubed_attributes=cubed_attributes,
        attribute_combinations=compute_datacube_groupings(cubed_attributes),
        metric_attributes_list=metric_attributes_list,
        metric_attributes=",".join(metric_attributes_list),
        custom_distribution_metadata_list=custom_dist_metadata,
        **kwargs,
    )


def probe_counts(**kwargs):
    """Variables for probe counts."""
    attributes = ["ping_type", "os", "app_version", "app_build_id", "channel"]

    return dict(
        attributes=",".join(attributes),
        attributes_no_os=",".join([attr for attr in attributes if attr != "os"]),
        aggregate_attributes="""
            metric,
            metric_type,
            key
        """,
        aggregate_grouping="""
            client_agg_type,
            agg_type
        """,
        # not boolean
        scalar_metric_types="""
            "counter",
            "quantity",
            "labeled_counter",
            "timespan"
        """,
        boolean_metric_types="""
            "boolean"
        """,
        **kwargs,
    )


def user_counts(**kwargs):
    """Variables for user counts."""
    attributes = ["ping_type", "os", "app_version", "app_build_id", "channel"]
    fixed_attributes = ["app_version", "channel"]
    cubed_attributes = [x for x in attributes if x not in fixed_attributes]

    return dict(
        attributes=",".join(attributes),
        cubed_attributes=cubed_attributes,
        attribute_combinations=compute_datacube_groupings(cubed_attributes),
        **kwargs,
    )


def sample_counts(**kwargs):
    """Variables for sample counts."""
    attributes = ["ping_type", "os", "app_version", "app_build_id", "channel"]
    fixed_attributes = ["app_version", "channel"]
    cubed_attributes = [x for x in attributes if x not in fixed_attributes]
    return dict(
        attributes=",".join(attributes),
        cubed_attributes=cubed_attributes,
        attribute_combinations=compute_datacube_groupings(cubed_attributes),
        **kwargs,
    )



"""
Dry run query files.

Passes all queries to a Cloud Function that will run the
queries with the dry_run option enabled.

We could provision BigQuery credentials to the CircleCI job to allow it to run
the queries directly, but there is no way to restrict permissions such that
only dry runs can be performed. In order to reduce risk of CI or local users
accidentally running queries during tests and overwriting production data, we
proxy the queries through the dry run service endpoint.
"""

import glob
import json
import re
import sys
from enum import Enum
from os.path import basename, dirname, exists
from pathlib import Path
from typing import Optional, Set
from urllib.request import Request, urlopen

import click
import google.auth
from google.auth.transport.requests import Request as GoogleAuthRequest
from google.cloud import bigquery
from google.oauth2.id_token import fetch_id_token

from .config import ConfigLoader
from .metadata.parse_metadata import Metadata
from .util.common import render

try:
    from functools import cached_property  # type: ignore
except ImportError:
    # python 3.7 compatibility
    from backports.cached_property import cached_property  # type: ignore


def get_credentials(auth_req: Optional[GoogleAuthRequest] = None):
    """Get GCP credentials."""
    auth_req = auth_req or GoogleAuthRequest()
    credentials, _ = google.auth.default(
        scopes=["https://www.googleapis.com/auth/cloud-platform"]
    )
    credentials.refresh(auth_req)
    return credentials


def get_id_token(dry_run_url=ConfigLoader.get("dry_run", "function"), credentials=None):
    """Get token to authenticate against Cloud Function."""
    auth_req = GoogleAuthRequest()
    credentials = credentials or get_credentials(auth_req)

    if hasattr(credentials, "id_token"):
        # Get token from default credentials for the current environment created via Cloud SDK run
        id_token = credentials.id_token
    else:
        # If the environment variable GOOGLE_APPLICATION_CREDENTIALS is set to service account JSON file,
        # then ID token is acquired using this service account credentials.
        id_token = fetch_id_token(auth_req, dry_run_url)
    return id_token


class Errors(Enum):
    """DryRun errors that require special handling."""

    READ_ONLY = 1
    DATE_FILTER_NEEDED = 2
    DATE_FILTER_NEEDED_AND_SYNTAX = 3


class DryRun:
    """Dry run SQL files."""

    def __init__(
        self,
        sqlfile,
        content=None,
        strip_dml=False,
        use_cloud_function=True,
        client=None,
        respect_skip=True,
        sql_dir=ConfigLoader.get("default", "sql_dir"),
        id_token=None,
        credentials=None,
        project=None,
        dataset=None,
        table=None,
    ):
        """Instantiate DryRun class."""
        self.sqlfile = sqlfile
        self.content = content
        self.strip_dml = strip_dml
        self.use_cloud_function = use_cloud_function
        self.bq_client = client
        self.respect_skip = respect_skip
        self.dry_run_url = ConfigLoader.get("dry_run", "function")
        self.sql_dir = sql_dir
        self.id_token = (
            id_token
            if not use_cloud_function or id_token
            else get_id_token(self.dry_run_url)
        )
        self.credentials = credentials
        self.project = project
        self.dataset = dataset
        self.table = table
        try:
            self.metadata = Metadata.of_query_file(self.sqlfile)
        except FileNotFoundError:
            self.metadata = None

        from bigquery_etl.cli.utils import is_authenticated

        if not is_authenticated():
            print(
                "Authentication to GCP required. Run `gcloud auth login  --update-adc` "
                "and check that the project is set correctly."
            )
            sys.exit(1)

    @cached_property
    def client(self):
        """Get BigQuery client instance."""
        if self.use_cloud_function:
            return None
        return self.bq_client or bigquery.Client(credentials=self.credentials)

    @staticmethod
    def skipped_files(sql_dir=ConfigLoader.get("default", "sql_dir")) -> Set[str]:
        """Return files skipped by dry run."""
        default_sql_dir = Path(ConfigLoader.get("default", "sql_dir"))
        sql_dir = Path(sql_dir)
        file_pattern_re = re.compile(rf"^{re.escape(str(default_sql_dir))}/")

        skip_files = {
            file
            for skip in ConfigLoader.get("dry_run", "skip", fallback=[])
            for file in glob.glob(
                file_pattern_re.sub(f"{str(sql_dir)}/", skip),
                recursive=True,
            )
        }

        # update skip list to include renamed queries in stage.
        test_project = ConfigLoader.get("default", "test_project", fallback="")
        file_pattern_re = re.compile(r"sql/([^\/]+)/([^/]+)(/?.*|$)")

        skip_files.update(
            [
                file
                for skip in ConfigLoader.get("dry_run", "skip", fallback=[])
                for file in glob.glob(
                    file_pattern_re.sub(
                        lambda x: f"sql/{test_project}/{x.group(2)}_{x.group(1).replace('-', '_')}*{x.group(3)}",
                        skip,
                    ),
                    recursive=True,
                )
            ]
        )

        return skip_files

    def skip(self):
        """Determine if dry run should be skipped."""
        return self.respect_skip and self.sqlfile in self.skipped_files(
            sql_dir=self.sql_dir
        )

    def get_sql(self):
        """Get SQL content."""
        if exists(self.sqlfile):
            file_path = Path(self.sqlfile)
            sql = render(
                file_path.name,
                format=False,
                template_folder=file_path.parent.absolute(),
            )
        else:
            raise ValueError(f"Invalid file path: {self.sqlfile}")
        if self.strip_dml:
            sql = re.sub(
                "CREATE OR REPLACE VIEW.*?AS",
                "",
                sql,
                flags=re.DOTALL,
            )
            sql = re.sub(
                "CREATE MATERIALIZED VIEW.*?AS",
                "",
                sql,
                flags=re.DOTALL,
            )

        return sql

    @cached_property
    def dry_run_result(self):
        """Dry run the provided SQL file."""
        if self.content:
            sql = self.content
        else:
            sql = self.get_sql()
        if self.metadata:
            # use metadata to rewrite date-type query params as submission_date
            date_params = [
                query_param
                for query_param in (
                    self.metadata.scheduling.get("date_partition_parameter"),
                    *(
                        param.split(":", 1)[0]
                        for param in self.metadata.scheduling.get("parameters", [])
                        if re.fullmatch(r"[^:]+:DATE:{{.*ds.*}}", param)
                    ),
                )
                if query_param and query_param != "submission_date"
            ]
            if date_params:
                pattern = re.compile(
                    "@("
                    + "|".join(date_params)
                    # match whole query parameter names
                    + ")(?![a-zA-Z0-9_])"
                )
                sql = pattern.sub("@submission_date", sql)
        project = basename(dirname(dirname(dirname(self.sqlfile))))
        dataset = basename(dirname(dirname(self.sqlfile)))
        try:
            if self.use_cloud_function:
                json_data = {
                    "project": self.project or project,
                    "dataset": self.dataset or dataset,
                    "query": sql,
                }

                if self.table:
                    json_data["table"] = self.table

                r = urlopen(
                    Request(
                        self.dry_run_url,
                        headers={
                            "Content-Type": "application/json",
                            "Authorization": f"Bearer {self.id_token}",
                        },
                        data=json.dumps(json_data).encode("utf8"),
                        method="POST",
                    )
                )
                return json.load(r)
            else:
                self.client.project = project
                job_config = bigquery.QueryJobConfig(
                    dry_run=True,
                    use_query_cache=False,
                    default_dataset=f"{project}.{dataset}",
                    query_parameters=[
                        bigquery.ScalarQueryParameter(
                            "submission_date", "DATE", "2019-01-01"
                        )
                    ],
                )
                job = self.client.query(sql, job_config=job_config)
                try:
                    dataset_labels = self.client.get_dataset(job.default_dataset).labels
                except Exception as e:
                    # Most users do not have bigquery.datasets.get permission in
                    # moz-fx-data-shared-prod
                    # This should not prevent the dry run from running since the dataset
                    # labels are usually not required
                    if "Permission bigquery.datasets.get denied on dataset" in str(e):
                        dataset_labels = []
                    else:
                        raise e

                result = {
                    "valid": True,
                    "referencedTables": [
                        ref.to_api_repr() for ref in job.referenced_tables
                    ],
                    "schema": (
                        job._properties.get("statistics", {})
                        .get("query", {})
                        .get("schema", {})
                    ),
                    "datasetLabels": dataset_labels,
                }
                if (
                    self.project is not None
                    and self.table is not None
                    and self.dataset is not None
                ):
                    table = self.client.get_table(
                        f"{self.project}.{self.dataset}.{self.table}"
                    )
                    result["tableMetadata"] = {
                        "tableType": table.table_type,
                        "friendlyName": table.friendly_name,
                        "schema": {
                            "fields": [field.to_api_repr() for field in table.schema]
                        },
                    }

                return result

        except Exception as e:
            print(f"{self.sqlfile!s:59} ERROR\n", e)
            return None

    def get_referenced_tables(self):
        """Return referenced tables by dry running the SQL file."""
        if not self.skip() and not self.is_valid():
            raise Exception(f"Error when dry running SQL file {self.sqlfile}")

        if self.skip():
            print(f"\t...Ignoring dryrun results for {self.sqlfile}")

        if (
            self.dry_run_result
            and self.dry_run_result["valid"]
            and "referencedTables" in self.dry_run_result
        ):
            return self.dry_run_result["referencedTables"]

        # Handle views that require a date filter
        if (
            self.dry_run_result
            and self.strip_dml
            and self.get_error() == Errors.DATE_FILTER_NEEDED
        ):
            # Since different queries require different partition filters
            # (submission_date, crash_date, timestamp, submission_timestamp, ...)
            # We can extract the filter name from the error message
            # (by capturing the next word after "column(s)")

            # Example error:
            # "Cannot query over table <table_name> without a filter over column(s)
            # <date_filter_name> that can be used for partition elimination."

            error = self.dry_run_result["errors"][0].get("message", "")
            date_filter = find_next_word("column(s)", error)

            if "date" in date_filter:
                filtered_content = (
                    f"{self.get_sql()}WHERE {date_filter} > current_date()"
                )
                if (
                    DryRun(
                        self.sqlfile,
                        filtered_content,
                        client=self.client,
                        id_token=self.id_token,
                    ).get_error()
                    == Errors.DATE_FILTER_NEEDED_AND_SYNTAX
                ):
                    # If the date filter (e.g. WHERE crash_date > current_date())
                    # is added to a query that already has a WHERE clause,
                    # it will throw an error. To fix this, we need to
                    # append 'AND' instead of 'WHERE'
                    filtered_content = (
                        f"{self.get_sql()}AND {date_filter} > current_date()"
                    )

            if "timestamp" in date_filter:
                filtered_content = (
                    f"{self.get_sql()}WHERE {date_filter} > current_timestamp()"
                )
                if (
                    DryRun(
                        sqlfile=self.sqlfile,
                        content=filtered_content,
                        client=self.client,
                        id_token=self.id_token,
                    ).get_error()
                    == Errors.DATE_FILTER_NEEDED_AND_SYNTAX
                ):
                    filtered_content = (
                        f"{self.get_sql()}AND {date_filter} > current_timestamp()"
                    )

            stripped_dml_result = DryRun(
                sqlfile=self.sqlfile,
                content=filtered_content,
                client=self.client,
                id_token=self.id_token,
            )
            if (
                stripped_dml_result.get_error() is None
                and "referencedTables" in stripped_dml_result.dry_run_result
            ):
                return stripped_dml_result.dry_run_result["referencedTables"]

        return []

    def get_schema(self):
        """Return the query schema by dry running the SQL file."""
        if not self.skip() and not self.is_valid():
            raise Exception(f"Error when dry running SQL file {self.sqlfile}")

        if self.skip():
            print(f"\t...Ignoring dryrun results for {self.sqlfile}")
            return {}

        if (
            self.dry_run_result
            and self.dry_run_result["valid"]
            and "schema" in self.dry_run_result
        ):
            return self.dry_run_result["schema"]

        return {}

    def get_table_schema(self):
        """Return the schema of the provided table."""
        if not self.skip() and not self.is_valid():
            raise Exception(f"Error when dry running SQL file {self.sqlfile}")

        if self.skip():
            print(f"\t...Ignoring dryrun results for {self.sqlfile}")
            return {}

        if (
            self.dry_run_result
            and self.dry_run_result["valid"]
            and "tableMetadata" in self.dry_run_result
        ):
            return self.dry_run_result["tableMetadata"]["schema"]

        return []

    def get_dataset_labels(self):
        """Return the labels on the default dataset by dry running the SQL file."""
        if not self.skip() and not self.is_valid():
            raise Exception(f"Error when dry running SQL file {self.sqlfile}")

        if self.skip():
            print(f"\t...Ignoring dryrun results for {self.sqlfile}")
            return {}

        if (
            self.dry_run_result
            and self.dry_run_result["valid"]
            and "datasetLabels" in self.dry_run_result
        ):
            return self.dry_run_result["datasetLabels"]

        return {}

    def is_valid(self):
        """Dry run the provided SQL file and check if valid."""
        if self.dry_run_result is None:
            return False

        if self.dry_run_result["valid"]:
            print(f"{self.sqlfile!s:59} OK")
        elif self.get_error() == Errors.READ_ONLY:
            # We want the dryrun service to only have read permissions, so
            # we expect CREATE VIEW and CREATE TABLE to throw specific
            # exceptions.
            print(f"{self.sqlfile!s:59} OK")
        elif self.get_error() == Errors.DATE_FILTER_NEEDED and self.strip_dml:
            # With strip_dml flag, some queries require a partition filter
            # (submission_date, submission_timestamp, etc.) to run
            # We mark these requests as valid and add a date filter
            # in get_referenced_table()
            print(f"{self.sqlfile!s:59} OK but DATE FILTER NEEDED")
        else:
            print(f"{self.sqlfile!s:59} ERROR\n", self.dry_run_result["errors"])
            return False

        return True

    def errors(self):
        """Dry run the provided SQL file and return errors."""
        if self.dry_run_result is None:
            return []
        return self.dry_run_result.get("errors", [])

    def get_error(self) -> Optional[Errors]:
        """Get specific errors for edge case handling."""
        errors = self.errors()
        if len(errors) != 1:
            return None

        error = errors[0]
        if error and error.get("code") in [400, 403]:
            error_message = error.get("message", "")
            if (
                "does not have bigquery.tables.create permission for dataset"
                in error_message
                or "Permission bigquery.tables.create denied" in error_message
                or "Permission bigquery.datasets.update denied" in error_message
            ):
                return Errors.READ_ONLY
            if "without a filter over column(s)" in error_message:
                return Errors.DATE_FILTER_NEEDED
            if (
                "Syntax error: Expected end of input but got keyword WHERE"
                in error_message
            ):
                return Errors.DATE_FILTER_NEEDED_AND_SYNTAX
        return None

    def validate_schema(self):
        """Check whether schema is valid."""
        # delay import to prevent circular imports in 'bigquery_etl.schema'
        from .schema import SCHEMA_FILE, Schema

        if (
            self.skip()
            or basename(self.sqlfile) == "script.sql"
            or str(self.sqlfile).endswith(".py")
        ):  # noqa E501
            print(f"\t...Ignoring schema validation for {self.sqlfile}")
            return True

        query_file_path = Path(self.sqlfile)
        query_schema = Schema.from_json(self.get_schema())
        if self.errors():
            # ignore file when there are errors that self.get_schema() did not raise
            click.echo(f"\t...Ignoring schema validation for {self.sqlfile}")
            return True
        existing_schema_path = query_file_path.parent / SCHEMA_FILE

        if not existing_schema_path.is_file():
            click.echo(f"No schema file defined for {query_file_path}", err=True)
            return True

        table_name = query_file_path.parent.name
        dataset_name = query_file_path.parent.parent.name
        project_name = query_file_path.parent.parent.parent.name

        partitioned_by = None
        if (
            self.metadata
            and self.metadata.bigquery
            and self.metadata.bigquery.time_partitioning
        ):
            partitioned_by = self.metadata.bigquery.time_partitioning.field

        table_schema = Schema.for_table(
            project_name,
            dataset_name,
            table_name,
            client=self.client,
            id_token=self.id_token,
            partitioned_by=partitioned_by,
        )

        # This check relies on the new schema being deployed to prod
        if not query_schema.compatible(table_schema):
            click.echo(
                click.style(
                    f"ERROR: Schema for query in {query_file_path} "
                    f"incompatible with schema deployed for "
                    f"{project_name}.{dataset_name}.{table_name}\n"
                    f"Did you deploy new the schema to prod yet?",
                    fg="red",
                ),
                err=True,
            )
            return False
        else:
            existing_schema = Schema.from_schema_file(existing_schema_path)

            if not existing_schema.equal(query_schema):
                click.echo(
                    click.style(
                        f"ERROR: Schema defined in {existing_schema_path} "
                        f"incompatible with query {query_file_path}",
                        fg="red",
                    ),
                    err=True,
                )
                return False

        click.echo(f"Schemas for {query_file_path} are valid.")
        return True


def sql_file_valid(sqlfile):
    """Dry run SQL files."""
    return DryRun(sqlfile).is_valid()


def find_next_word(target, source):
    """Find the next word in a string."""
    split = source.split()
    for i, w in enumerate(split):
        if w == target:
            # get the next word, and remove quotations from column name
            return split[i + 1].replace("'", "")



"""Testing data for query."""

from pathlib import Path
from uuid import uuid4

import yaml

ROOT = Path(__file__).parent

SUBMISSION_DATE = "2020-10-01"
APP_BUILD_ID = "2020100100"

# needs > 5 clients to be considered a version
CLIENTS_DAILY_SCALAR_AGGREGATES = [
    {
        "submission_date": SUBMISSION_DATE,
        "client_id": str(uuid4()),
        "ping_type": "metrics",
        "os": "Android",
        "app_version": 84,
        "app_build_id": APP_BUILD_ID,
        "channel": "nightly",
        "scalar_aggregates": [],
    }
    for _ in range(6)
]

EXPECT = [dict(channel="nightly", latest_version=84)]

prefix = "glam_etl"
tables = [
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__"
        "view_clients_daily_scalar_aggregates_v1.yaml",
        CLIENTS_DAILY_SCALAR_AGGREGATES,
    ),
    ("expect.yaml", EXPECT),
]
for name, data in tables:
    with (ROOT / name).open("w") as fp:
        yaml.dump(data, fp)



"""microsoft_store data - download deliverables, clean and upload to BigQuery."""

import csv
import json
import os
import re
import tempfile
from argparse import ArgumentParser
from time import sleep

import requests
from google.cloud import bigquery

API_URI = "https://manage.devcenter.microsoft.com"

"""MICROSOFT_STORE_APP_LIST is a list of dictionaries.
Keys are the app_name, and app_tenant_id in the form:
{"app_name":"<specific_app_name from Microsoft Store", "tenant_id":"<unique Microsoft tenant_id"}
Look here to see the apps we are tracking: https://partner.microsoft.com/en-us/dashboard/insights/analytics/reports/summary
"""
CSV_FIELDS = [
    "date",
    "application_id",
    "application_name",
    "custom_campaign_id",
    "referrer_uri_domain",
    "channel_type",
    "store_client",
    "device_type",
    "market",
    "click_count",
    "conversion_count",
]

MS_CLIENT_ID = os.environ.get("MICROSOFT_CLIENT_ID")
MS_CLIENT_SECRET = os.environ.get("MICROSOFT_CLIENT_SECRET")
MS_APP_LIST = os.environ.get("MICROSOFT_STORE_APP_LIST")
MS_TENANT_ID = os.environ.get("MICROSOFT_TENANT_ID")


def post_response(url, headers, data):
    """POST response function."""
    response = requests.post(url, headers=headers, data=data)
    if (response.status_code == 401) or (response.status_code == 400):
        print(f"***Error: {response.status_code}***")
        print(response.text)
    return response


def get_response(url, headers, params):
    """GET response function."""
    response = requests.get(url, headers=headers, params=params)
    if (response.status_code == 401) or (response.status_code == 400):
        print(f"***Error: {response.status_code}***")
        print(response.text)
    return response


def read_json(filename: str) -> dict:
    """Read JSON file."""
    with open(filename, "r") as f:
        print(f.read())
        data = json.loads(f.read())
    return data


def change_null_to_string(json_string):
    """Change null values in downloaded json string to string."""
    none_string = re.sub("null", '"null"', json_string)
    return none_string


def write_dict_to_csv(json_data, filename):
    """Write a dictionary to a csv."""
    with open(filename, "w") as out_file:
        dict_writer = csv.DictWriter(out_file, CSV_FIELDS)
        dict_writer.writeheader()
        dict_writer.writerows(json_data)


def microsoft_authorization(tenant_id, client_id, client_secret, resource_url):
    """Microsoft Store Authoriazation. Returns the bearer token required for data download."""
    url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/token"
    query_params = {
        "grant_type": "client_credentials",
        "client_id": client_id,
        "client_secret": client_secret,
        "resource": resource_url,
    }
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    json_file = post_response(url, headers, query_params)
    response_data = json.loads(json_file.text)
    bearer_token = response_data["access_token"]
    return bearer_token


def download_microsoft_store_data(date, application_id, bearer_token):
    """Download data from Microsoft - application_id, bearer_token are called here."""
    start_date = date  #
    end_date = date
    token = bearer_token
    app_id = application_id
    # getting overview metrics for different kpis / Deliverables
    url = f"https://manage.devcenter.microsoft.com/v1.0/my/analytics/appchannelconversions?applicationId={app_id}"
    url_params = (
        f"aggregationLevel=day&startDate={start_date}&endDate={end_date}&skip=0"
    )
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Authorization": f"Bearer {token}",
    }
    print(url)
    response = get_response(url, headers, url_params)
    return response


def clean_json(query_export, date):
    """Turn the json file into a list to be input into a CSV for bq upload."""
    fields_list = []
    for val in query_export["Value"]:
        field_dict = {
            "date": date,
            "application_id": val["applicationId"],
            "application_name": val["applicationName"],
            "custom_campaign_id": val["customCampaignId"],
            "referrer_uri_domain": val["referrerUriDomain"],
            "channel_type": val["channelType"],
            "store_client": val["storeClient"],
            "device_type": val["deviceType"],
            "market": val["market"],
            "click_count": val["clickCount"],
            "conversion_count": val["conversionCount"],
        }
        fields_list.append(field_dict)
    return fields_list


def upload_to_bigquery(csv_data, project, dataset, table_name, date):
    """Upload the data to bigquery."""
    date = date
    print("writing json to csv")
    partition = f"{date}".replace("-", "")
    print(partition)
    with tempfile.NamedTemporaryFile() as tmp_csv:
        with open(tmp_csv.name, "w+b") as f_csv:
            write_dict_to_csv(csv_data, f_csv.name)
            client = bigquery.Client(project)
            job_config = bigquery.LoadJobConfig(
                create_disposition="CREATE_IF_NEEDED",
                write_disposition="WRITE_TRUNCATE",
                time_partitioning=bigquery.TimePartitioning(
                    type_=bigquery.TimePartitioningType.DAY,
                    field="date",
                ),
                skip_leading_rows=1,
                schema=[
                    bigquery.SchemaField("date", "DATE"),
                    bigquery.SchemaField("application_id", "STRING"),
                    bigquery.SchemaField("application_name", "STRING"),
                    bigquery.SchemaField("custom_campaign_id", "STRING"),
                    bigquery.SchemaField("referrer_uri_domain", "STRING"),
                    bigquery.SchemaField("channel_type", "STRING"),
                    bigquery.SchemaField("store_client", "STRING"),
                    bigquery.SchemaField("device_type", "STRING"),
                    bigquery.SchemaField("market", "STRING"),
                    bigquery.SchemaField("click_count", "INT64"),
                    bigquery.SchemaField("conversion_count", "INT64"),
                ],
            )
            destination = f"{project}.{dataset}.{table_name}${partition}"

            job = client.load_table_from_file(f_csv, destination, job_config=job_config)
            print(
                f"Writing microsoft_store data for all apps to {destination}. BigQuery job ID: {job.job_id}"
            )
            job.result()


def main():
    """Input data, call functions, get stuff done."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("--date", required=True)
    parser.add_argument("--project", default="moz-fx-data-shared-prod")
    parser.add_argument("--dataset", default="microsoft_derived")

    args = parser.parse_args()

    project = args.project
    dataset = args.dataset
    table_name = "app_conversions_v1"

    date = args.date
    client_id = MS_CLIENT_ID
    client_secret = MS_CLIENT_SECRET
    app_list = MS_APP_LIST
    tenant_id = MS_TENANT_ID
    resource_url = API_URI

    ms_app_list = json.loads(app_list)

    data = []

    bearer_token = microsoft_authorization(
        tenant_id, client_id, client_secret, resource_url
    )

    # Cycle through the apps to get the relevant data
    for app in ms_app_list:
        print(f'This is data for {app["app_name"]} - {app["app_id"]} ')
        # Ping the microsoft_store URL and get a response
        json_file = download_microsoft_store_data(date, app["app_id"], bearer_token)
        # For some reason, the date returned from https://manage.devcenter.microsoft.com/v1.0/my/analytics/appchannelconversions? returns the date as null.
        # This needs to be changed from a null to a string null.
        json_none_string = change_null_to_string(json_file.text)
        # Convert the string to a dictionary
        query_export = eval(json_none_string)
        if query_export["Value"]:
            # This section writes the tmp json data into a temp CSV file which will then be put into a BigQuery table
            microsoft_store_data = clean_json(query_export, date)
            data.extend(microsoft_store_data)
        else:
            print("no data for today")
        sleep(5)
    upload_to_bigquery(data, project, dataset, table_name, date)


if __name__ == "__main__":
    main()



"""
Determine BigQuery table storage timeline per day.

To read more on the source table, please visit:
https://cloud.google.com/bigquery/docs/information-schema-table-storage-timeline
"""

from argparse import ArgumentParser

from google.cloud import bigquery

DEFAULT_PROJECTS = [
    "mozdata",
    "moz-fx-data-shared-prod",
    "moz-fx-data-marketing-prod",
]

parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-shared-prod")
# projects queries were run from that access table
parser.add_argument("--source_projects", nargs="+", default=DEFAULT_PROJECTS)
parser.add_argument("--destination_dataset", default="monitoring_derived")
parser.add_argument(
    "--destination_table", default="bigquery_table_storage_timeline_daily_v1"
)


def create_query(date, source_project):
    """Create query for a source project.  1GB = POW(1024, 3) bytes."""
    return f"""
        SELECT
          DATE('{date}') AS change_date,
          project_id,
          table_schema AS dataset_id,
          table_name AS table_id,
          deleted,
          DATE(creation_time) AS creation_date,
          count(*) AS change_count,
          avg(total_rows) AS avg_total_rows,
          avg(total_partitions) AS avg_total_partitions,
          avg(total_logical_bytes) AS avg_total_logical_bytes,
          avg(active_logical_bytes) AS avg_active_logical_bytes,
          avg(long_term_logical_bytes) AS avg_long_term_logical_bytes,
          avg(total_physical_bytes) AS avg_total_physical_bytes,
          avg(active_physical_bytes) AS avg_active_physical_bytes,
          avg(long_term_physical_bytes) AS avg_long_term_physical_bytes,
          avg(time_travel_physical_bytes) AS avg_time_travel_physical_bytes
        FROM `{source_project}.region-us.INFORMATION_SCHEMA.TABLE_STORAGE_TIMELINE`
        WHERE
          DATE(timestamp) = '{date}'
        GROUP BY change_date, project_id, dataset_id, table_id, deleted, creation_date
        ORDER BY change_date, project_id, dataset_id, table_id, deleted, creation_date
    """


def main():
    """Run query for each source project."""
    args = parser.parse_args()

    partition = args.date.replace("-", "")
    destination_table = f"{args.project}.{args.destination_dataset}.{args.destination_table}${partition}"

    # remove old partition in case of re-run
    client = bigquery.Client(args.project)
    client.delete_table(destination_table, not_found_ok=True)

    for project in args.source_projects:
        client = bigquery.Client(project)
        query = create_query(args.date, project)
        job_config = bigquery.QueryJobConfig(
            destination=destination_table, write_disposition="WRITE_APPEND"
        )
        client.query(query, job_config=job_config).result()


if __name__ == "__main__":
    main()



"""
Copy a day's data from live to stable ping tables, deduplicating on document_id.

By default, the script will process all tables in datasets named
like *_live, copying data into table of the same name in *_stable
datasets. The script can be configured to exclude a list of tables
or to process only a specific list of tables.
"""

import json
import logging
from datetime import datetime, timedelta
from functools import partial
from itertools import groupby
from multiprocessing.pool import ThreadPool

import click
from google.api_core.exceptions import BadRequest
from google.cloud import bigquery

from bigquery_etl.cli.utils import table_matches_patterns
from bigquery_etl.util.bigquery_id import sql_table_id
from bigquery_etl.util.client_queue import ClientQueue
from bigquery_etl.util.common import TempDatasetReference

from .cli.utils import parallelism_option, project_id_option

QUERY_TEMPLATE = """
WITH
  -- Distinct document_ids and their minimum submission_timestamp today
  -- not including document_ids that only occur on or after @end_time
  distinct_document_ids AS (
  SELECT
    document_id,
    MIN(submission_timestamp) AS submission_timestamp
  FROM
    `{live_table}`
  WHERE
    DATE(submission_timestamp) >= DATE_SUB(
        DATE(@start_time),
        INTERVAL @num_preceding_days DAY
    )
    AND submission_timestamp < @end_time
    -- Bug 1657360
    AND 'automation' NOT IN (
      SELECT TRIM(t) FROM UNNEST(SPLIT(metadata.header.x_source_tags, ',')) t
    )
  GROUP BY
    document_id
  HAVING
    submission_timestamp >= @start_time),
  -- A single slice of a live ping table.
  base AS (
  SELECT
    *
  FROM
    `{live_table}`
  JOIN
    distinct_document_ids
    -- Retain only the first seen documents for each ID, according to timestamp.
    USING (document_id, submission_timestamp)
  WHERE
    submission_timestamp >= @start_time
    AND submission_timestamp < @end_time),
  --
  -- Order documents by assigning a row number.
  numbered_duplicates AS (
  SELECT
    *,
    ROW_NUMBER() OVER (PARTITION BY document_id) AS _n
  FROM
    base)
  --
  -- Retain only one document for each ID.
SELECT
  * EXCEPT(_n)
FROM
  numbered_duplicates
WHERE
  _n = 1
"""


def _get_query_job_configs(
    client,
    live_table,
    date,
    dry_run,
    slices,
    priority,
    preceding_days,
    num_retries,
    temp_dataset,
):
    sql = QUERY_TEMPLATE.format(live_table=live_table)
    stable_table = f"{live_table.replace('_live.', '_stable.', 1)}${date:%Y%m%d}"
    kwargs = dict(use_legacy_sql=False, dry_run=dry_run, priority=priority)
    start_time = datetime(*date.timetuple()[:6])
    end_time = start_time + timedelta(days=1)
    if slices > 1:
        # create temporary tables with stable_table's time_partitioning and
        # clustering_fields, and a 1-day expiration
        stable_table = client.get_table(stable_table)
        ddl = "CREATE TABLE\n  `{dest}`"
        ddl += f"\nPARTITION BY\n  DATE({stable_table.time_partitioning.field})"
        if stable_table.clustering_fields:
            ddl += f"\nCLUSTER BY\n  {', '.join(stable_table.clustering_fields)}"
        ddl += (
            "\nOPTIONS"
            "\n  ("
            "\n    partition_expiration_days = CAST('inf' AS FLOAT64),"
            "\n    expiration_timestamp = "
            "TIMESTAMP_ADD(CURRENT_TIMESTAMP, INTERVAL 1 DAY)"
            "\n  )"
        )
        slice_size = (end_time - start_time) / slices
        params = [start_time + slice_size * i for i in range(slices)] + [
            end_time
        ]  # explicitly use end_time to avoid rounding errors
        return [
            (
                f"{ddl.format(dest=temp_dataset.temp_table())}\nAS\n{sql.strip()}",
                stable_table,
                bigquery.QueryJobConfig(
                    query_parameters=[
                        bigquery.ScalarQueryParameter(
                            "start_time", "TIMESTAMP", params[i]
                        ),
                        bigquery.ScalarQueryParameter(
                            "end_time", "TIMESTAMP", params[i + 1]
                        ),
                        bigquery.ScalarQueryParameter(
                            "num_preceding_days", "INT64", preceding_days
                        ),
                    ],
                    **kwargs,
                ),
                num_retries,
            )
            for i in range(slices)
        ]
    else:
        return [
            (
                sql,
                stable_table,
                bigquery.QueryJobConfig(
                    destination=stable_table,
                    write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
                    query_parameters=[
                        bigquery.ScalarQueryParameter(
                            "start_time", "TIMESTAMP", start_time
                        ),
                        bigquery.ScalarQueryParameter(
                            "end_time", "TIMESTAMP", end_time
                        ),
                        bigquery.ScalarQueryParameter(
                            "num_preceding_days", "INT64", preceding_days
                        ),
                    ],
                    **kwargs,
                ),
                num_retries,
            )
        ]


def _run_deduplication_query(client, sql, stable_table, job_config, num_retries):
    query_job = client.query(sql, job_config, job_id_prefix="copy_dedup_")
    if not query_job.dry_run:
        try:
            query_job.result()
        except BadRequest as e:
            if num_retries <= 0:
                raise
            logging.warn("Encountered bad request, retrying: ", e)
            return _run_deduplication_query(
                client, sql, stable_table, job_config, num_retries - 1
            )
    logging.info(
        f"Completed query job {query_job.job_id} for {stable_table}"
        f" with params: {job_config.query_parameters}"
    )
    return stable_table, query_job


def _copy_join_parts(client, stable_table, query_jobs):
    total_bytes = sum(query.total_bytes_processed for query in query_jobs)
    if query_jobs[0].dry_run:
        api_repr = json.dumps(query_jobs[0].to_api_repr())
        if len(query_jobs) > 1:
            logging.info(f"Would process {total_bytes} bytes: [{api_repr},...]")
            logging.info(f"Would copy {len(query_jobs)} results to {stable_table}")
        else:
            logging.info(f"Would process {total_bytes} bytes: {api_repr}")
    else:
        total_slot_hours = round(
            sum(query.slot_millis for query in query_jobs) / 1000 / 60 / 60, 3
        )
        logging.info(
            f"Processed {total_bytes} bytes in {total_slot_hours}"
            f" slot hours to populate {stable_table}"
        )
        if len(query_jobs) > 1:
            partition_id = stable_table.table_id.split("$", 1)[1]
            sources = [
                f"{sql_table_id(job.destination)}${partition_id}" for job in query_jobs
            ]
            job_config = bigquery.CopyJobConfig(
                write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE
            )
            copy_job = client.copy_table(sources, stable_table, job_config=job_config)
            copy_job.result()
            logging.info(f"Copied {len(query_jobs)} results to populate {stable_table}")
            for job in query_jobs:
                client.delete_table(job.destination)
            logging.info(f"Deleted {len(query_jobs)} temporary tables")


def _contains_glob(patterns):
    return any(set("*?[").intersection(pattern) for pattern in patterns)


def _glob_dataset(pattern):
    return pattern.split(".", 1)[0]


def _list_live_tables(client, pool, project_id, only_tables, table_filter):
    if only_tables and not _contains_glob(only_tables):
        # skip list calls when only_tables exists and contains no globs
        return [f"{project_id}.{t}" for t in only_tables if table_filter(t)]
    if only_tables and not _contains_glob(_glob_dataset(t) for t in only_tables):
        # skip list_datasets call when only_tables exists and datasets contain no globs
        live_datasets = {f"{project_id}.{_glob_dataset(t)}" for t in only_tables}
    else:
        live_datasets = [
            d.reference
            for d in client.list_datasets(project_id)
            if d.dataset_id.endswith("_live")
        ]
    return [
        sql_table_id(t)
        for tables in pool.map(client.list_tables, live_datasets)
        for t in tables
        if table_filter(f"{t.dataset_id}.{t.table_id}")
        and "beam_load_sink" not in t.table_id
    ]


@click.command(
    "copy_deduplicate",
    help="Copy a day's data from live to stable ping tables, deduplicating on document_id",
)
@project_id_option("moz-fx-data-shar-nonprod-efed")
@click.option(
    "--dates",
    "--date",
    multiple=True,
    required=True,
    type=lambda d: datetime.strptime(d, "%Y-%m-%d").date(),
    help="One or more days of data to copy, in format 2019-01-01",
)
@parallelism_option()
@click.option(
    "--dry_run",
    "--dry-run",
    is_flag=True,
    help="Do not make changes, only log actions that would be taken",
)
@click.option(
    "--log-level",
    "--log_level",
    help="Log level.",
    default=logging.getLevelName(logging.INFO),
    type=str.upper,
)
@click.option(
    "--priority",
    default=bigquery.QueryPriority.INTERACTIVE,
    type=click.Choice(
        [bigquery.QueryPriority.BATCH, bigquery.QueryPriority.INTERACTIVE],
        case_sensitive=False,
    ),
    help="Priority for BigQuery query jobs; BATCH priority may significantly slow "
    "down queries if reserved slots are not enabled for the billing project; "
    "INTERACTIVE priority is limited to 100 concurrent queries per project",
)
@click.option(
    "--temp-dataset",
    "--temp_dataset",
    "--temporary-dataset",
    "--temporary_dataset",
    default="moz-fx-data-shared-prod.tmp",
    type=TempDatasetReference.from_string,
    help="Dataset where intermediate query results will be temporarily stored, "
    "formatted as PROJECT_ID.DATASET_ID",
)
@click.option(
    "--slices",
    type=int,
    default=1,
    help=(
        "Number of queries to split deduplicate over, each handling an equal-size time "
        "slice of the date; avoids memory overflow at the cost of less effective "
        "clustering; recommended only for tables failing due to memory overflow"
    ),
)
@click.option(
    "--hourly",
    is_flag=True,
    help="Deduplicate one hour at a time; equivalent to --slices=24",
)
@click.option(
    "--preceding_days",
    "--preceding-days",
    type=int,
    default=0,
    help="Number of days preceding --date that should be used to filter out duplicates",
)
@click.option(
    "--num_retries",
    "--num-retries",
    type=int,
    default=2,
    help="Number of times to retry each slice in case of query error",
)
@click.option(
    "--billing-projects",
    "--billing_projects",
    "--billing-project",
    "--billing_project",
    "-p",
    multiple=True,
    default=[None],
    help="One or more billing projects over which bigquery jobs should be "
    "distributed",
)
@click.option(
    "--except",
    "-x",
    "exclude",
    multiple=True,
    help="Process all tables except for the given tables",
)
@click.option(
    "--only",
    "-o",
    multiple=True,
    help="Process only the given tables",
)
def copy_deduplicate(
    project_id,
    dates,
    parallelism,
    dry_run,
    log_level,
    priority,
    temp_dataset,
    slices,
    hourly,
    preceding_days,
    num_retries,
    billing_projects,
    exclude,
    only,
):
    """Copy a day's data from live to stable ping tables, dedup on document_id."""
    # create a queue for balancing load across projects
    client_q = ClientQueue(billing_projects, parallelism)

    if hourly:
        slices = 24

    table_filter = partial(table_matches_patterns, "*", False)

    if only:
        table_filter = partial(table_matches_patterns, list(only), False)
    elif exclude:
        table_filter = partial(table_matches_patterns, list(exclude), True)

    with ThreadPool(parallelism) as pool:
        with client_q.client() as client:
            live_tables = _list_live_tables(
                client=client,
                pool=pool,
                project_id=project_id,
                only_tables=only,
                table_filter=table_filter,
            )

            query_jobs = [
                (_run_deduplication_query, *args)
                for jobs in pool.starmap(
                    _get_query_job_configs,
                    [
                        (
                            client,  # only use one client to create temp tables
                            live_table,
                            date,
                            dry_run,
                            slices,
                            priority,
                            preceding_days,
                            num_retries,
                            temp_dataset,
                        )
                        for live_table in live_tables
                        for date in dates
                    ],
                )
                for args in jobs
            ]

        # preserve query_jobs order so results stay sorted by stable_table for groupby
        results = pool.starmap(client_q.with_client, query_jobs, chunksize=1)
        copy_jobs = [
            (_copy_join_parts, stable_table, [query_job for _, query_job in group])
            for stable_table, group in groupby(results, key=lambda result: result[0])
        ]
        pool.starmap(client_q.with_client, copy_jobs, chunksize=1)



"""BigQuery ETL."""

from pathlib import Path

from .config import ConfigLoader

# We need to set the project dir before this module is imported from cli.__init__.py
if __name__ == "bigquery_etl":
    ConfigLoader.set_project_dir(Path().absolute())



"""microsoft_store data - download deliverables, clean and upload to BigQuery."""

import csv
import json
import os
import tempfile
from argparse import ArgumentParser
from time import sleep

import requests
from google.cloud import bigquery

API_URI = "https://manage.devcenter.microsoft.com"

"""MICROSOFT_STORE_APP_LIST is a list of dictionaries.
Keys are the app_name, and app_tenant_id in the form:
{"app_name":"<specific_app_name from Microsoft Store", "tenant_id":"<unique Microsoft tenant_id"}
Look here to see the apps we are tracking: https://partner.microsoft.com/en-us/dashboard/insights/analytics/reports/summary
"""
CSV_FIELDS = [
    "date",
    "application_id",
    "application_name",
    "package_version",
    "device_type",
    "market",
    "os_version",
    "successful_install_count",
]

MS_CLIENT_ID = os.environ.get("MICROSOFT_CLIENT_ID")
MS_CLIENT_SECRET = os.environ.get("MICROSOFT_CLIENT_SECRET")
MS_APP_LIST = os.environ.get("MICROSOFT_STORE_APP_LIST")
MS_TENANT_ID = os.environ.get("MICROSOFT_TENANT_ID")


def post_response(url, headers, data):
    """POST response function."""
    response = requests.post(url, headers=headers, data=data)
    if (response.status_code == 401) or (response.status_code == 400):
        print(f"***Error: {response.status_code}***")
        print(response.text)
    return response


def get_response(url, headers, params):
    """GET response function."""
    response = requests.get(url, headers=headers, params=params)
    if (response.status_code == 401) or (response.status_code == 400):
        print(f"***Error: {response.status_code}***")
        print(response.text)
    return response


def read_json(filename: str) -> dict:
    """Read JSON file."""
    with open(filename, "r") as f:
        data = json.loads(f.read())
    return data


def write_dict_to_csv(json_data, filename):
    """Write a dictionary to a csv."""
    with open(filename, "w") as out_file:
        dict_writer = csv.DictWriter(out_file, CSV_FIELDS)
        dict_writer.writeheader()
        dict_writer.writerows(json_data)


def microsoft_authorization(tenant_id, client_id, client_secret, resource_url):
    """Microsoft Store Authoriazation. Returns the bearer token required for data download."""
    url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/token"
    query_params = {
        "grant_type": "client_credentials",
        "client_id": client_id,
        "client_secret": client_secret,
        "resource": resource_url,
    }
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    json_file = post_response(url, headers, query_params)
    response_data = json.loads(json_file.text)
    bearer_token = response_data["access_token"]
    return bearer_token


def download_microsoft_store_data(date, application_id, bearer_token):
    """Download data from Microsoft - application_id, bearer_token are called here."""
    start_date = date  #
    end_date = date
    token = bearer_token
    app_id = application_id
    groupBy = [
        "applicationName",
        "date",
        "deviceType",
        "market",
        "osVersion",
    ]
    # getting overview metrics for different kpis / Deliverables
    url = f"https://manage.devcenter.microsoft.com/v1.0/my/analytics/installs?applicationId={app_id}"
    url_params = f"aggregationLevel=day&startDate={start_date}&endDate={end_date}&skip=0&groupBy={','.join(groupBy)}&filter=date eq '{date}'&orderby=date"
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Authorization": f"Bearer {token}",
    }
    print(url)
    response = get_response(url, headers, url_params)
    return response


def check_json(microsoft_store_response_text):
    """Script will return an empty dictionary for apps on days when there is no data. Check for that here."""
    with tempfile.NamedTemporaryFile() as tmp_json:
        with open(tmp_json.name, "w") as f_json:
            f_json.write(microsoft_store_response_text)
            try:
                query_export = read_json(f_json.name)
            except (
                ValueError
            ):  # ex. json.decoder.JSONDecodeError: Expecting value: line 1 column 1 (char 0)
                return None
    return query_export


def clean_json(query_export, date):
    """Turn the json file into a list to be input into a CSV for bq upload."""
    fields_list = []
    for val in query_export["Value"]:
        field_dict = {
            "date": val["date"],
            "application_id": val["applicationId"],
            "application_name": val["applicationName"],
            "package_version": val["packageVersion"],
            "device_type": val["deviceType"],
            "market": val["market"],
            "os_version": val["osVersion"],
            "successful_install_count": val["successfulInstallCount"],
        }
        fields_list.append(field_dict)
    return fields_list


def upload_to_bigquery(csv_data, project, dataset, table_name, date):
    """Upload the data to bigquery."""
    print("writing json to csv")
    partition = f"{date}".replace("-", "")
    print(partition)
    with tempfile.NamedTemporaryFile() as tmp_csv:
        with open(tmp_csv.name, "w+b") as f_csv:
            write_dict_to_csv(csv_data, f_csv.name)
            client = bigquery.Client(project)
            job_config = bigquery.LoadJobConfig(
                create_disposition="CREATE_IF_NEEDED",
                write_disposition="WRITE_TRUNCATE",
                time_partitioning=bigquery.TimePartitioning(
                    type_=bigquery.TimePartitioningType.DAY,
                    field="date",
                ),
                skip_leading_rows=1,
                schema=[
                    bigquery.SchemaField("date", "DATE"),
                    bigquery.SchemaField("application_id", "STRING"),
                    bigquery.SchemaField("application_name", "STRING"),
                    bigquery.SchemaField("package_version", "STRING"),
                    bigquery.SchemaField("device_type", "STRING"),
                    bigquery.SchemaField("market", "STRING"),
                    bigquery.SchemaField("os_version", "STRING"),
                    bigquery.SchemaField("successful_install_count", "INT64"),
                ],
            )
            destination = f"{project}.{dataset}.{table_name}${partition}"

            job = client.load_table_from_file(f_csv, destination, job_config=job_config)
            print(
                f"Writing microsoft_store data for all apps to {destination}. BigQuery job ID: {job.job_id}"
            )
            job.result()


def main():
    """Input data, call functions, get stuff done."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("--date", required=True)
    parser.add_argument("--project", default="moz-fx-data-shared-prod")
    parser.add_argument("--dataset", default="microsoft_derived")

    args = parser.parse_args()

    project = args.project
    dataset = args.dataset
    table_name = "app_installs_v1"

    date = args.date
    client_id = MS_CLIENT_ID
    client_secret = MS_CLIENT_SECRET
    app_list = MS_APP_LIST
    tenant_id = MS_TENANT_ID
    resource_url = API_URI

    ms_app_list = json.loads(app_list)

    data = []

    bearer_token = microsoft_authorization(
        tenant_id, client_id, client_secret, resource_url
    )

    # Cycle through the apps to get the relevant data
    for app in ms_app_list:
        print(f'This is data for {app["app_name"]} - {app["app_id"]} ')
        # Ping the microsoft_store URL and get a response
        json_file = download_microsoft_store_data(date, app["app_id"], bearer_token)
        query_export = check_json(json_file.text)
        if query_export is not None:
            # This section writes the tmp json data into a temp CSV file which will then be put into a BigQuery table
            microsoft_store_data = clean_json(query_export, date)
            data.extend(microsoft_store_data)
        else:
            print("no data for today")
        sleep(5)
    upload_to_bigquery(data, project, dataset, table_name, date)


if __name__ == "__main__":
    main()



from bigquery_etl.dependency import extract_table_references


class TestDependency:
    def test_extract_table_refs_correctly_ignores_unpivot(self):
        unpivot_query = "SELECT * FROM a UNPIVOT(b FOR c IN (d, e, f))"
        refs = extract_table_references(unpivot_query)

        assert refs == ["a"]

    def test_extract_table_refs_correctly_ignores_pivot(self):
        pivot_query = """SELECT * FROM Produce
          PIVOT(SUM(sales) FOR quarter IN ('Q1', 'Q2', 'Q3', 'Q4'))
        """
        refs = extract_table_references(pivot_query)

        assert refs == ["Produce"]

    def test_extract_table_refs_pivot_and_join(self):
        pivot_query = """SELECT * FROM Produce
          PIVOT(SUM(sales) FOR quarter IN ('Q1', 'Q2', 'Q3', 'Q4'))
          JOIN Perishable_Mints USING (name)
        """
        refs = extract_table_references(pivot_query)

        assert set(refs) == {"Produce", "Perishable_Mints"}



"""Build and use query dependency graphs."""

import re
import sys
from glob import glob
from itertools import groupby
from pathlib import Path
from subprocess import CalledProcessError
from typing import Dict, Iterator, List, Tuple

import rich_click as click
import sqlglot
import yaml

from bigquery_etl.config import ConfigLoader
from bigquery_etl.schema.stable_table_schema import get_stable_table_schemas
from bigquery_etl.util.common import render

stable_views = None


def _raw_table_name(table: sqlglot.exp.Table) -> str:
    with_replacements = (
        table.sql("bigquery", comments=False)
        # remove alias
        .split(" AS ", 1)[0]
        # remove quotes
        .replace("`", "")
    )
    # remove PIVOT/UNPIVOT
    removed_pivots = re.sub(" (?:UN)?PIVOT.*$", "", with_replacements)

    return removed_pivots


def extract_table_references(sql: str) -> List[str]:
    """Return a list of tables referenced in the given SQL."""
    # sqlglot cannot handle scripts with variables and control statements
    if re.search(r"^\s*DECLARE\b", sql, flags=re.MULTILINE):
        return []
    # sqlglot parses UDFs with keyword names incorrectly:
    # https://github.com/tobymao/sqlglot/issues/3332
    sql = re.sub(
        r"\.(true|false|null)\(",
        r".`\1`(",
        sql,
        flags=re.IGNORECASE,
    )
    query = sqlglot.parse(sql, read="bigquery")
    creates = set()
    tables = set()
    for statement in query:
        if statement is None:
            continue
        creates |= {
            _raw_table_name(expr.this)
            for expr in statement.find_all(sqlglot.exp.Create)
        }
        tables |= (
            {_raw_table_name(table) for table in statement.find_all(sqlglot.exp.Table)}
            # ignore references created in this query
            - creates
            # ignore CTEs created in this statement
            - {cte.alias_or_name for cte in statement.find_all(sqlglot.exp.CTE)}
        )
    return sorted(tables)


def extract_table_references_without_views(path: Path) -> Iterator[str]:
    """Recursively search for non-view tables referenced in the given SQL file."""
    global stable_views

    sql = render(path.name, template_folder=path.parent)
    for table in extract_table_references(sql):
        ref_base = path.parent
        parts = tuple(table.split("."))
        for _ in parts:
            ref_base = ref_base.parent
        view_paths = [ref_base.joinpath(*parts, "view.sql")]
        if parts[:1] == (
            ConfigLoader.get("default", "user_facing_project", fallback="mozdata"),
        ):
            view_paths.append(
                ref_base.joinpath(
                    ConfigLoader.get(
                        "default", "project", fallback="moz-fx-data-shared-prod"
                    ),
                    *parts[1:],
                    "view.sql",
                ),
            )
        for view_path in view_paths:
            if view_path == path:
                continue  # skip self references
            if view_path.is_file():
                yield from extract_table_references_without_views(view_path)
                break
        else:
            # use directory structure to fully qualify table names
            while len(parts) < 3:
                parts = (ref_base.name, *parts)
                ref_base = ref_base.parent
            if parts[:-2] in (
                (
                    ConfigLoader.get(
                        "default", "project", fallback="moz-fx-data-shared-prod"
                    ),
                ),
                (
                    ConfigLoader.get(
                        "default", "user_facing_project", fallback="mozdata"
                    ),
                ),
            ):
                if stable_views is None:
                    # lazy read stable views
                    stable_views = {
                        tuple(schema.user_facing_view.split(".")): tuple(
                            schema.stable_table.split(".")
                        )
                        for schema in get_stable_table_schemas()
                    }
                if parts[-2:] in stable_views:
                    parts = (
                        ConfigLoader.get(
                            "default", "project", fallback="moz-fx-data-shared-prod"
                        ),
                        *stable_views[parts[-2:]],
                    )
            yield ".".join(parts)


def _get_references(
    paths: Tuple[str, ...], without_views: bool = False
) -> Iterator[Tuple[Path, List[str]]]:
    file_paths = {
        path
        for parent in map(Path, paths or ["sql"])
        for path in (
            map(Path, glob(f"{parent}/**/*.sql", recursive=True))
            if parent.is_dir()
            else [parent]
        )
        if not path.name.endswith(".template.sql")  # skip templates
    }
    fail = False
    for path in sorted(file_paths):
        try:
            if without_views:
                yield path, list(extract_table_references_without_views(path))
            else:
                sql = render(path.name, template_folder=path.parent)
                yield path, extract_table_references(sql)
        except CalledProcessError as e:
            raise click.ClickException(f"failed to import jnius: {e}")
        except ImportError as e:
            raise click.ClickException(*e.args)
        except ValueError as e:
            fail = True
            print(f"Failed to parse file {path}: {e}", file=sys.stderr)

    if fail:
        raise click.ClickException("Some paths could not be analyzed")


def get_dependency_graph(
    paths: Tuple[str, ...], without_views: bool = False
) -> Dict[str, List[str]]:
    """Return the query dependency graph."""
    refs = _get_references(paths, without_views=without_views)
    dependency_graph = {}

    for ref in refs:
        table = ref[0].parent.name
        dataset = ref[0].parent.parent.name
        project = ref[0].parent.parent.parent.name
        dependency_graph[f"{project}.{dataset}.{table}"] = ref[1]

    return dependency_graph


@click.group(help=__doc__)
def dependency():
    """Create the CLI group for dependency commands."""
    pass


@dependency.command(
    help="Show table references in sql files.",
)
@click.argument(
    "paths",
    nargs=-1,
    type=click.Path(file_okay=True),
)
@click.option(
    "--without-views",
    "--without_views",
    is_flag=True,
    help="recursively resolve view references to underlying tables",
)
def show(paths: Tuple[str, ...], without_views: bool):
    """Show table references in sql files."""
    for path, table_references in _get_references(paths, without_views):
        if table_references:
            for table in table_references:
                print(f"{path}: {table}")
        else:
            print(f"{path} contains no table references", file=sys.stderr)


@dependency.command(
    help="Record table references in metadata. Fails if metadata already contains "
    "references section.",
)
@click.argument(
    "paths",
    nargs=-1,
    type=click.Path(file_okay=True),
)
@click.option(
    "--skip-existing",
    "--skip_existing",
    is_flag=True,
    help="Skip files with existing references rather than failing",
)
def record(paths: Tuple[str, ...], skip_existing):
    """Record table references in metadata."""
    for parent, group in groupby(_get_references(paths), lambda e: e[0].parent):
        references = {
            path.name: table_references
            for path, table_references in group
            if table_references
        }
        if not references:
            continue
        with open(parent / "metadata.yaml", "a+") as f:
            f.seek(0)
            metadata = yaml.safe_load(f)
            if metadata is None:
                pass  # new or empty metadata.yaml
            elif not isinstance(metadata, dict):
                raise click.ClickException(f"{f.name} is not valid metadata")
            elif "references" in metadata:
                if skip_existing:
                    # Continue without modifying metadata.yaml
                    continue
                raise click.ClickException(f"{f.name} already contains references")
            f.write("\n# Generated by bigquery_etl.dependency\n")
            f.write(yaml.dump({"references": references}))



"""Methods for publishing static CSV files to BigQuery."""

import functools
import glob
import logging
import os
from pathlib import Path

import rich_click as click
from google.cloud import bigquery
from google.cloud.bigquery import DatasetReference

from bigquery_etl.cli.utils import project_id_option
from bigquery_etl.config import ConfigLoader
from bigquery_etl.metadata.parse_metadata import (
    DATASET_METADATA_FILE,
    METADATA_FILE,
    DatasetMetadata,
    Metadata,
)
from bigquery_etl.metadata.publish_metadata import publish_metadata
from bigquery_etl.schema import SCHEMA_FILE, Schema
from bigquery_etl.util.common import project_dirs

DATA_FILENAME = "data.csv"


@click.group("static", help="Commands for working with static CSV files.")
def static_():
    """Create the CLI group for static commands."""
    pass


@static_.command("publish", help="Publish CSV files as BigQuery tables.")
@project_id_option()
def publish(project_id):
    """Publish CSV files as BigQuery tables."""
    source_project = project_id
    target_project = project_id
    if target_project == ConfigLoader.get(
        "default", "user_facing_project", fallback="mozdata"
    ):
        source_project = ConfigLoader.get(
            "default", "project", fallback="moz-fx-data-shared-prod"
        )

    for project_dir in project_dirs(source_project):
        # Assumes directory structure is project/dataset/table/files.
        for data_file_path in glob.iglob(
            os.path.join(project_dir, "*", "*", DATA_FILENAME)
        ):
            table_dir = os.path.dirname(data_file_path)
            dataset_dir = os.path.dirname(table_dir)
            if target_project == ConfigLoader.get(
                "default", "user_facing_project", fallback="mozdata"
            ) and not _is_user_facing_dataset(dataset_dir):
                logging.debug(
                    f"Skipping `{data_file_path}` for {target_project} because it isn't"
                    " in a user-facing dataset."
                )
                continue

            schema_file_path = os.path.join(table_dir, SCHEMA_FILE)
            if not os.path.exists(schema_file_path):
                schema_file_path = None

            metadata_file_path = os.path.join(table_dir, METADATA_FILE)
            if not os.path.exists(metadata_file_path):
                metadata_file_path = None

            _load_table(
                data_file_path,
                schema_file_path,
                metadata_file_path,
                target_project,
            )


def _load_table(
    data_file_path, schema_file_path=None, metadata_file_path=None, project=None
):
    # Assume path is ...project/dataset/table/data.csv
    path_split = os.path.normcase(data_file_path).split(os.path.sep)
    dataset_id = path_split[-3]
    table_id = path_split[-2]
    if not project:
        project = path_split[-4]
    logging.info(
        f"Loading `{project}.{dataset_id}.{table_id}` table from `{data_file_path}`."
    )

    client = bigquery.Client(project)
    dataset_ref = DatasetReference(project, dataset_id)
    table_ref = dataset_ref.table(table_id)

    job_config = bigquery.LoadJobConfig(
        source_format=bigquery.SourceFormat.CSV,
        skip_leading_rows=1,
        allow_quoted_newlines=True,
        write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
    )
    with open(data_file_path, "rb") as data_file:
        if schema_file_path is None:
            fields = data_file.readline().decode().strip().split(",")
            # Assume all fields are strings and nullable
            job_config.schema = [
                bigquery.SchemaField(field, field_type="STRING") for field in fields
            ]
            data_file.seek(0)
        else:
            job_config.schema = Schema.from_schema_file(
                Path(schema_file_path)
            ).to_bigquery_schema()

        job = client.load_table_from_file(data_file, table_ref, job_config=job_config)

    job.result()

    if metadata_file_path is not None:
        metadata = Metadata.from_file(metadata_file_path)
        publish_metadata(client, project, dataset_id, table_id, metadata)


@functools.lru_cache
def _is_user_facing_dataset(dataset_directory):
    dataset_metadata_file_path = os.path.join(dataset_directory, DATASET_METADATA_FILE)
    return (
        os.path.exists(dataset_metadata_file_path)
        and DatasetMetadata.from_file(dataset_metadata_file_path).user_facing
    )



#!/usr/bin/env python3
"""Longitudinal query generator."""
import argparse
import datetime
import sys
import textwrap


def parse_date(date_str):
    """Read dates in YYYMMDD format."""
    return datetime.datetime.strptime(date_str, "%Y%m%d")


def comma_separated(names):
    """Parse a comma-separated list of names."""
    return [x.strip() for x in names.split(",")]


p = argparse.ArgumentParser()
p.add_argument("--tablename", type=str, help="Table to pull data from", required=True)
p.add_argument(
    "--from",
    type=parse_date,
    help="From submission date. Defaults to 6 months before `to`.",
)
p.add_argument("--to", type=parse_date, help="To submission date", required=True)
p.add_argument(
    "--submission-date-col",
    type=str,
    help="Name of the submission date column. Defaults to submission_date_s3",
    default="submission_date_s3",
)
p.add_argument(
    "--select",
    type=str,
    help="Select statement to retrieve data with; e.g. "
    '"substr(subsession_start_date, 0, 10) as activity_date". If none given, '
    "defaults to all input columns",
    default="*",
)
p.add_argument(
    "--where",
    type=str,
    help="Where SQL clause, filtering input data. E.g. "
    "'normalized_channel = \"nightly\"'",
)
p.add_argument(
    "--grouping-column",
    type=str,
    help="Column to group data by. Defaults to client_id",
    default="client_id",
)
p.add_argument(
    "--ordering-columns",
    type=comma_separated,
    help="Columns to order the arrays by (comma separated). "
    "Defaults to submission-date-col",
)
p.add_argument(
    "--max-array-length",
    type=int,
    help="Max length of any groups history. Defaults to no limit. "
    "Negatives are ignored",
)


def six_months_before(d):
    """Calculate a date six months before the given date."""
    m = d.month - 6
    y = d.year
    if m < 0:
        m += 12
        y -= 1
    return d.replace(month=m, year=y)


def generate_sql(opts):
    """Create a BigQuery SQL query for collecting a longitudinal dataset."""
    opts.update(
        {
            "from": datetime.datetime.strftime(
                opts["from"] or six_months_before(opts["to"]), "'%Y-%m-%d'"
            ),
            "to": datetime.datetime.strftime(opts["to"], "'%Y-%m-%d'"),
            "where": (
                "\n{}AND {}".format(" " * 10, opts["where"]) if opts["where"] else ""
            ),
            "ordering_columns": ", ".join(
                opts["ordering_columns"] or [opts["submission_date_col"]]
            ),
            "limit": opts["max_array_length"] or "UNBOUNDED",
        }
    )
    if opts["grouping_column"] in opts["ordering_columns"]:
        raise ValueError(
            "{!r} can't be used as both a grouping and ordering column".format(
                opts["grouping_column"], opts["ordering_columns"]
            )
        )

    return textwrap.dedent(
        """
      WITH aggregated AS (
        SELECT
         ROW_NUMBER() OVER (PARTITION BY {grouping_column} ORDER BY {ordering_columns}) AS _n,
         {grouping_column},
         ARRAY_AGG(STRUCT({select})) OVER (PARTITION BY {grouping_column} ORDER BY {ordering_columns} ROWS BETWEEN CURRENT ROW AND {limit} FOLLOWING) AS aggcol
        FROM
          {tablename}
        WHERE
          {submission_date_col} > {from}
          AND {submission_date_col} <= {to}{where})
      SELECT
        * except (_n)
      FROM
        aggregated
      WHERE
        _n = 1
        """.format(  # noqa: E501
            **opts
        )
    )


def main(argv, out=print):
    """Print a longitudinal query to stdout."""
    opts = p.parse_args(argv[1:])
    out(generate_sql(vars(opts)))


if __name__ == "__main__":
    main(sys.argv)



"""Monitor the number of rows deleted by shredder."""

import datetime
from argparse import ArgumentParser
from multiprocessing.pool import ThreadPool
from pathlib import Path

import yaml
from google.cloud import bigquery

from bigquery_etl.util import standard_args

PROJECT_IDS_QUERY = """
SELECT DISTINCT
  SPLIT(job_id, ":")[OFFSET(0)] AS project_id,
FROM
  `moz-fx-data-shredder.shredder_state.shredder_state`
WHERE
  DATE(job_created) BETWEEN @end_date - 1 AND @end_date
"""

JOB_INFO_QUERY = """
WITH shredder AS (
  SELECT DISTINCT
    job_id AS full_job_id,
    SPLIT(job_id, ":")[OFFSET(0)] AS project_id,
    SPLIT(job_id, ".")[OFFSET(1)] AS job_id,
  FROM
    `moz-fx-data-shredder.shredder_state.shredder_state`
  WHERE
    DATE(job_created) BETWEEN @end_date - 1 AND @end_date
    -- ignore intermediate sample id queries; only use final copy job
    AND NOT REGEXP_CONTAINS(task_id, "__sample_[0-9]{{1,2}}$")
),
successful_jobs AS (
  -- https://cloud.google.com/bigquery/docs/information-schema-jobs
  SELECT
    *
  FROM
    `{project_id}.region-us.INFORMATION_SCHEMA.JOBS_BY_PROJECT`
  WHERE
    DATE(creation_time) BETWEEN @end_date - 1 AND @end_date
    AND DATE(end_time) = @end_date
    AND state = "DONE"
    AND error_result IS NULL
)
SELECT
  successful_jobs.end_time,
  shredder.full_job_id AS job_id,
  successful_jobs.destination_table.project_id,
  successful_jobs.destination_table.dataset_id,
  successful_jobs.destination_table.table_id,
  successful_jobs.dml_statistics.deleted_row_count,
  UNIX_MILLIS(successful_jobs.start_time) AS start_time_millis,
  UNIX_MILLIS(successful_jobs.end_time) AS end_time_millis,
FROM
  shredder
JOIN
  successful_jobs
  USING (project_id, job_id)
"""

parser = ArgumentParser()
parser.add_argument(
    "--end_date",
    "--end-date",
    type=datetime.date.fromisoformat,
    help="",
)
parser.add_argument(
    "--destination_table",
    "--destination-table",
    help="Table where results will be written",
)
standard_args.add_parallelism(parser)


def get_job_info(
    client: bigquery.Client, project_id: str, end_date: datetime.date
) -> list[dict]:
    query = client.query(
        JOB_INFO_QUERY.format(project_id=project_id),
        job_config=bigquery.QueryJobConfig(
            query_parameters=[
                bigquery.ScalarQueryParameter("end_date", "DATE", end_date),
            ]
        ),
    )
    return list(map(dict, query.result()))


def add_job_metadata(client: bigquery.Client, job: dict) -> dict:
    if job["deleted_row_count"] is None:
        table = f'{job["project_id"]}.{job["dataset_id"]}.{job["table_id"]}'
        before = client.get_table(f'{table}@{job["start_time_millis"]}').num_rows
        after = client.get_table(f'{table}@{job["end_time_millis"]}').num_rows
        job["deleted_row_count"] = count = before - after
        if count < 0:
            raise ValueError(f"deleted_row_count must be >= 0, but got: {count}")
    # move partition id to separate column
    if "$" in job["table_id"]:
        job["table_id"], _, job["partition_id"] = job["table_id"].partition("$")
    job["end_time"] = job["end_time"].isoformat()
    return job


def main():
    args = parser.parse_args()

    client = bigquery.Client()

    query = client.query(
        PROJECT_IDS_QUERY,
        job_config=bigquery.QueryJobConfig(
            query_parameters=[
                bigquery.ScalarQueryParameter("end_date", "DATE", args.end_date),
            ]
        ),
    )
    project_ids = [row["project_id"] for row in query.result()]

    with ThreadPool(args.parallelism) as pool:
        jobs = sum(
            pool.starmap(
                get_job_info,
                ((client, project_id, args.end_date) for project_id in project_ids),
                chunksize=1,
            ),
            start=[],
        )
        json_rows = list(
            pool.starmap(
                add_job_metadata, ((client, job) for job in jobs), chunksize=1
            ),
        )

    load_job = client.load_table_from_json(
        json_rows=json_rows,
        destination=args.destination_table,
        job_config=bigquery.LoadJobConfig(
            ignore_unknown_values=True,
            schema=yaml.safe_load((Path(__file__).parent / "schema.yaml").read_text())[
                "fields"
            ],
            source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
            time_partitioning=bigquery.TimePartitioning(field="end_time"),
            write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
        ),
    )
    load_job.result()


if __name__ == "__main__":
    main()



import os
from pathlib import Path

from generate_queries import TemplatedDir, get_query_dirs

BASE_DIR = Path(os.path.dirname(__file__)).parent


class TestGenerateQueries:
    def test_get_query_dirs(self):
        res = list(get_query_dirs(BASE_DIR / "templates"))
        assert res == [
            TemplatedDir("event_types", BASE_DIR / "templates" / "event_types")
        ]



from click.testing import CliRunner

from bigquery_etl.cli.alchemer import backfill


def test_cli_alchemer_backfill_inclusive_dates(monkeypatch):
    def nop(*args, **kwargs):
        pass

    monkeypatch.setattr("bigquery_etl.cli.alchemer.get_survey_data", nop)
    monkeypatch.setattr("bigquery_etl.cli.alchemer.insert_to_bq", nop)

    result = CliRunner().invoke(
        backfill,
        [
            "--start-date",
            "2021-01-01",
            "--end-date",
            "2021-01-03",
            "--survey_id",
            "55555",
            "--api_token",
            "token",
            "--api_secret",
            "secret",
            "--destination_table",
            "testing_dataset.testing_table",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    # parse the output...
    assert (
        "2021-01-01" in result.stdout
        and "2021-01-02" in result.stdout
        and "2021-01-03" in result.stdout
    ), result.stdout



"""Generate and record schemas for user-facing derived dataset views."""

from functools import partial
from pathlib import Path

import click
from pathos.multiprocessing import ProcessingPool

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.dryrun import get_id_token

NON_USER_FACING_DATASET_SUBSTRINGS = (
    "_derived",
    "_external",
    "_bi",
    "_restricted",
    "udf",
)

VIEW_FILE = "view.sql"
METADATA_FILE = "metadata.yaml"
SCHEMA_FILE = "schema.yaml"


def _generate_view_schema(sql_dir, view_directory, id_token=None):
    import logging

    from bigquery_etl.dependency import extract_table_references
    from bigquery_etl.metadata.parse_metadata import Metadata
    from bigquery_etl.schema import Schema
    from bigquery_etl.util.common import render
    from bigquery_etl.view import View

    logging.basicConfig(format="%(levelname)s (%(filename)s:%(lineno)d) - %(message)s")

    # If the view references only one table, we can:
    # 1. Get the reference table partition column if it exists.
    #   (to dry run views to partitioned tables).
    # 2. Get the reference table schema and use it to enrich the
    #   view schema we get from dry-running.
    def _get_reference_dir_path(view_file):
        view_references = extract_table_references(
            render(view_file.name, view_file.parent)
        )
        if len(view_references) != 1:
            return

        target_project = view_file.parent.parent.parent.name
        target_dataset = view_file.parent.parent.name

        target_reference = view_references[0]
        parts = target_reference.split(".")
        if len(parts) == 3:
            reference_project_id, reference_dataset_id, reference_table_id = parts
        # Fully qualify the reference:
        elif len(parts) == 2:
            reference_project_id = target_project
            reference_dataset_id, reference_table_id = parts
        elif len(parts) == 1:
            reference_project_id = target_project
            reference_dataset_id = target_dataset
            reference_table_id = parts[0]
        else:
            return

        return (
            sql_dir / reference_project_id / reference_dataset_id / reference_table_id
        )

    def _get_reference_partition_column(ref_path):
        if ref_path is None:
            logging.debug("No table reference, skipping partition column.")
            return

        try:
            reference_metadata = Metadata.from_file(ref_path / METADATA_FILE)
        except Exception as metadata_exception:
            logging.warning(f"Unable to get reference metadata: {metadata_exception}")
            return

        bigquery_metadata = reference_metadata.bigquery
        if bigquery_metadata is None:
            logging.warning(
                f"No bigquery metadata at {ref_path}, unable to get partition column."
            )
            return

        partition_metadata = bigquery_metadata.time_partitioning
        if partition_metadata is None:
            logging.warning(
                f"No partition metadata at {ref_path}, unable to get partition column."
            )
            return

        return partition_metadata.field

    view_file = view_directory / VIEW_FILE
    if not view_file.exists():
        return

    reference_path = _get_reference_dir_path(view_file)

    # If this is a view to a stable table, don't try to write the schema:
    if reference_path is not None:
        reference_dataset = reference_path.parent.name
        if reference_dataset.endswith("_stable"):
            return

    # Optionally get the upstream partition column
    reference_partition_column = _get_reference_partition_column(reference_path)
    if reference_partition_column is None:
        logging.debug("No reference partition column, dry running without one.")

    view = View.from_file(
        view_file, partition_column=reference_partition_column, id_token=id_token
    )

    # `View.schema` prioritizes the configured schema over the dryrun schema, but here
    # we prioritize the dryrun schema because the `schema.yaml` file might be out of date.
    schema = view.dryrun_schema or view.configured_schema
    if view.dryrun_schema and view.configured_schema:
        try:
            schema.merge(
                view.configured_schema,
                attributes=["description"],
                add_missing_fields=False,
                ignore_missing_fields=True,
            )
        except Exception as e:
            logging.warning(
                f"Error enriching {view.view_identifier} view schema from {view.schema_path}: {e}"
            )
    if not schema:
        logging.warning(
            f"Couldn't get schema for {view.view_identifier} potentially "
            f"due to dry-run error. Won't write yaml."
        )
        return

    # Optionally enrich the view schema if we have a valid table reference
    if reference_path:
        reference_schema_file = reference_path / SCHEMA_FILE
        if reference_schema_file.exists():
            try:
                reference_schema = Schema.from_schema_file(reference_schema_file)
                schema.merge(
                    reference_schema,
                    attributes=["description"],
                    add_missing_fields=False,
                    ignore_missing_fields=True,
                )
            except Exception as e:
                logging.warning(
                    f"Error enriching {view.view_identifier} view schema from {reference_schema_file}: {e}"
                )

    schema.to_yaml_file(view_directory / SCHEMA_FILE)


@click.command("generate")
@click.option(
    "--target_project",
    "--target-project",
    help="Which project the views should be written to.",
    default="moz-fx-data-shared-prod",
)
@click.option(
    "--output-dir",
    "--output_dir",
    "--sql-dir",
    "--sql_dir",
    help="The location to write to. Defaults to sql/.",
    default=Path("sql"),
    type=click.Path(file_okay=False),
)
@click.option(
    "--parallelism",
    "-P",
    help="Maximum number of tasks to execute concurrently",
    default=20,
    type=int,
)
@use_cloud_function_option
def generate(target_project, output_dir, parallelism, use_cloud_function):
    """
    Generate schema yaml files for views in output_dir/target_project.

    We dry-run to get the schema data and where possible we enrich the
    view schemas with underlying table descriptions.
    """
    project_path = Path(f"{output_dir}/{target_project}")

    dataset_paths = [
        dataset_path
        for dataset_path in project_path.iterdir()
        if dataset_path.is_dir()
        and all(
            substring not in str(dataset_path)
            for substring in NON_USER_FACING_DATASET_SUBSTRINGS
        )
    ]

    id_token = get_id_token()
    view_directories = []

    for dataset_path in dataset_paths:
        view_directories += [
            path
            for path in dataset_path.iterdir()
            if path.is_dir() and (path / VIEW_FILE).exists()
        ]

    with ProcessingPool(parallelism) as pool:
        pool.map(
            partial(_generate_view_schema, Path(output_dir), id_token=id_token),
            view_directories,
        )



"""Format SQL."""

import re
from dataclasses import replace

from .tokenizer import (
    AliasSeparator,
    BlockEndKeyword,
    BlockKeyword,
    BlockStartKeyword,
    BuiltInFunctionIdentifier,
    CaseSubclause,
    ClosingBracket,
    Comment,
    ExpressionSeparator,
    FieldAccessOperator,
    Identifier,
    JinjaBlockEnd,
    JinjaBlockStart,
    JinjaBlockStatement,
    JinjaComment,
    JinjaExpression,
    JinjaStatement,
    LineComment,
    Literal,
    NewlineKeyword,
    OpeningBracket,
    Operator,
    ReservedKeyword,
    SpaceBeforeBracketKeyword,
    StatementSeparator,
    TopLevelKeyword,
    Whitespace,
    tokenize,
)


def simple_format(tokens, indent="  "):
    """Format tokens in a single pass."""
    first_token = True
    require_newline_before_next_token = False
    allow_space_before_next_bracket = False
    allow_space_before_next_token = False
    prev_was_block_end = False
    prev_was_jinja = False
    prev_was_statement_separator = False
    prev_was_unary_operator = False
    next_operator_is_unary = True
    indent_types = []
    can_format = True
    for token in tokens:
        # skip original whitespace tokens, unless formatting is disabled
        if isinstance(token, Whitespace):
            if not can_format:
                yield token
            continue

        # update state for current token
        if isinstance(token, Comment):
            # enable to disable formatting
            if can_format and token.format_off:
                can_format = False
            elif not can_format and token.format_on:
                can_format = True
        elif isinstance(token, ClosingBracket):
            # decrease indent to match last OpeningBracket
            while indent_types and indent_types.pop() is not OpeningBracket:
                pass
        elif isinstance(token, TopLevelKeyword):
            # decrease indent from previous TopLevelKeyword
            if indent_types and indent_types[-1] is TopLevelKeyword:
                indent_types.pop()
        elif isinstance(token, BlockEndKeyword):
            # decrease indent to match last BlockKeyword
            while indent_types and indent_types.pop() is not BlockKeyword:
                pass
            prev_was_statement_separator = False
        elif isinstance(token, JinjaBlockEnd):
            # decrease indent to match last JinjaBlockStart
            while indent_types and indent_types.pop() is not JinjaBlockStart:
                pass
        elif isinstance(token, CaseSubclause):
            if token.value.upper() in ("WHEN", "ELSE"):
                # Have WHEN and ELSE clauses indented one level more than CASE.
                while indent_types and indent_types[-1] is CaseSubclause:
                    indent_types.pop()
        elif isinstance(
            token, (AliasSeparator, ExpressionSeparator, FieldAccessOperator)
        ):
            if prev_was_block_end or prev_was_jinja:
                require_newline_before_next_token = False

        # yield whitespace
        if not can_format or isinstance(token, StatementSeparator) or first_token:
            # except between statements
            # no new whitespace when formatting is disabled
            # no space before statement separator
            # no space before first token
            pass
        elif isinstance(token, Comment):
            # blank line before comments if they start on their own line
            # and come after a statement separator, and before #fail and #warn
            if token.value.startswith("\n") and prev_was_statement_separator:
                yield Whitespace("\n")
            elif re.fullmatch(r"\s*#(fail|warn)", token.value):
                yield Whitespace("\n")
        elif (
            require_newline_before_next_token
            or isinstance(
                token,
                (NewlineKeyword, ClosingBracket, BlockKeyword, JinjaBlockStatement),
            )
            or prev_was_statement_separator
        ):
            if prev_was_statement_separator:
                yield Whitespace("\n")
            yield Whitespace("\n" + indent * len(indent_types))
        elif (
            allow_space_before_next_token
            and (
                allow_space_before_next_bracket or not isinstance(token, OpeningBracket)
            )
            and not isinstance(token, (FieldAccessOperator, ExpressionSeparator))
            and not (
                prev_was_unary_operator and isinstance(token, (Literal, Identifier))
            )
        ):
            yield Whitespace(" ")

        if can_format:
            # uppercase keywords and replace contained whitespace with single spaces
            if isinstance(token, ReservedKeyword):
                token = replace(token, value=re.sub(r"\s+", " ", token.value.upper()))
            # uppercase built-in function names
            elif isinstance(token, BuiltInFunctionIdentifier):
                token = replace(token, value=token.value.upper())

        yield token

        # update state for next token
        require_newline_before_next_token = isinstance(
            token,
            (
                Comment,
                BlockKeyword,
                TopLevelKeyword,
                OpeningBracket,
                ExpressionSeparator,
                StatementSeparator,
                JinjaStatement,
            ),
        )
        allow_space_before_next_token = not isinstance(token, FieldAccessOperator)
        prev_was_block_end = isinstance(token, BlockEndKeyword)
        prev_was_statement_separator = isinstance(token, StatementSeparator)
        prev_was_unary_operator = next_operator_is_unary and isinstance(token, Operator)
        prev_was_jinja = isinstance(
            token, (JinjaExpression, JinjaComment, JinjaStatement)
        )
        if not isinstance(token, Comment):
            # format next operator as unary if there is no preceding argument
            next_operator_is_unary = not isinstance(
                token, (Literal, Identifier, ClosingBracket)
            )
        allow_space_before_next_bracket = isinstance(
            token, (SpaceBeforeBracketKeyword, Operator)
        )
        if isinstance(token, TopLevelKeyword) and token.value == "WITH":
            # don't indent CTE's and don't put the first one on a new line
            require_newline_before_next_token = False
        elif isinstance(token, BlockStartKeyword):
            # increase indent
            indent_types.append(BlockKeyword)
        elif isinstance(token, JinjaBlockStart):
            # increase indent
            indent_types.append(JinjaBlockStart)
        elif isinstance(token, (TopLevelKeyword, OpeningBracket, CaseSubclause)):
            # increase indent
            indent_types.append(type(token))
        elif isinstance(token, StatementSeparator):
            # decrease for previous top level keyword
            if indent_types and indent_types[-1] is TopLevelKeyword:
                indent_types.pop()
        first_token = False


class Line:
    """Container for a line of tokens."""

    def __init__(self, indent_token=None, can_format=True):
        """Initialize."""
        self.indent_token = indent_token
        self.can_format = can_format and not isinstance(indent_token, Comment)
        if indent_token is None:
            self.indent_level = 0
        else:
            self.indent_level = len(indent_token.value)
            if indent_token.value.startswith("\n"):
                self.indent_level -= 1
        self.inline_tokens = []
        self.inline_length = 0

    def add(self, token):
        """Add a token to this line."""
        self.inline_length += len(token.value)
        self.inline_tokens.append(token)

    @property
    def tokens(self):
        """Get a list of all the tokens in this line."""
        if self.indent_token is None:
            return self.inline_tokens
        else:
            return [self.indent_token] + self.inline_tokens

    @property
    def can_start_inline_block(self):
        """Determine if this line starts a bracket block that may be inlined.

        This line starts a bracket block if it ends with an OpeningBracket.

        This line can't be inlined if formatting is disabled.

        Blocks preceded by an alias, such as common table expressions and
        window expressions, should not be inlined.
        """
        return (
            self.can_format
            and self.ends_with_opening_bracket
            and not (
                len(self.inline_tokens) > 2
                and isinstance(self.inline_tokens[-3], AliasSeparator)
            )
        )

    @property
    def ends_with_opening_bracket(self):
        """Determine if this line ends with an OpeningBracket."""
        return self.inline_tokens and isinstance(self.inline_tokens[-1], OpeningBracket)

    @property
    def starts_with_closing_bracket(self):
        """Determine if this line starts with a ClosingBracket."""
        return self.inline_tokens and isinstance(self.inline_tokens[0], ClosingBracket)

    @property
    def ends_with_line_comment(self):
        """Determine if this line ends with a line comment."""
        return self.inline_tokens and isinstance(self.inline_tokens[-1], LineComment)


def inline_block_format(tokens, max_line_length=100):
    """Extend simple_format to inline each bracket block if possible.

    A bracket block is a series of tokens from an opening bracket to the
    matching closing bracket. To inline a block means to put it on a single
    line, instead of multiple lines.

    Inline a block if the result would be shorter than max_line_length.

    Do not inline if block if contains a comment.

    For example, this formatter may convert:

        IF(
          condition,
          value_if_true,
          value_if_false,
        )

    to

        IF(condition, value_if_true, value_if_false)

    Implementation requires simple_format to put opening brackets at the end of
    the line and closing brackets at the beginning of the line, unless there is
    a comment between them.
    """
    # format tokens using simple_format, then group into lines
    lines = [Line()]
    can_format = True
    for token in simple_format(tokens):
        if token.value.startswith("\n"):
            lines.append(Line(token, can_format))
        else:
            lines[-1].add(token)
        if isinstance(token, Comment):
            if can_format and token.format_off:
                # disable formatting for current and following lines
                lines[-1].can_format = False
                can_format = False
            elif not can_format and token.format_on:
                # enable formatting for following lines
                can_format = True

    # combine all lines in each bracket block that fits in max_line_length
    skip_lines = 0
    for index, line in enumerate(lines):
        if skip_lines > 0:
            skip_lines -= 1
            continue
        yield from line.tokens
        if line.can_start_inline_block:
            indent_level = line.indent_level
            line_length = indent_level + line.inline_length
            pending_lines = 0
            pending = []
            open_brackets = 1
            index += 1  # start on the next line
            previous_line = line
            for line in lines[index:]:
                if not line.can_format:
                    break
                # Line comments can't be moved into the middle of a line.
                if previous_line.ends_with_line_comment:
                    break
                if (
                    not previous_line.ends_with_opening_bracket
                    and not line.starts_with_closing_bracket
                ):
                    pending.append(Whitespace(" "))
                    line_length += 1
                pending_lines += 1
                pending.extend(line.inline_tokens)
                line_length += line.inline_length
                if line_length > max_line_length:
                    break
                if line.starts_with_closing_bracket:
                    open_brackets -= 1
                if open_brackets == 0:
                    # flush pending and handle next block if present
                    yield from pending
                    skip_lines += pending_lines
                    if line.can_start_inline_block:
                        pending_lines = 0
                        pending = []
                    else:
                        break
                if line.ends_with_opening_bracket:
                    open_brackets += 1
                previous_line = line


def reformat(query, format_=inline_block_format, trailing_newline=False):
    """Reformat query and return as a string."""
    tokens = format_(tokenize(query))
    return "".join(token.value for token in tokens) + ("\n" if trailing_newline else "")



#!/usr/bin/env python

"""Retrieve and load daily stock index and foreign exchange data."""

import logging
import urllib
from argparse import ArgumentParser
from datetime import date, datetime
from typing import Any

import google.auth  # type: ignore
import requests
from google.cloud import bigquery

BASE_URL = "https://financialmodelingprep.com/api/v3"
INDEX_TICKERS = [
    "^DJI",  # Dow Jones Industrial Average
    "^GSPC",  # SNP - SNP Real Time Price. Currency in USD
    "^IXIC",  # Nasdaq GIDS - Nasdaq GIDS Real Time Price. Currency in USD
]
FOREX_TICKERS = [
    "EURUSD=X",  # Euro to USD exchange rate
    "GBPUSD=X",  # GB pound to USD exchange rate
]


parser = ArgumentParser(description=__doc__)
parser.add_argument("--api-key", required=True)
parser.add_argument("--destination_project", default="moz-fx-data-shared-prod")
parser.add_argument("--destination_dataset", default="reference_derived")
parser.add_argument("--destination_table", default="macroeconomic_indices_v1")
parser.add_argument("--market-date")
parser.add_argument("--backfill", action="store_true")
parser.add_argument("--start-date")
parser.add_argument("--end-date")


def get_index_ticker(
    api_key: str, ticker: str, start_date: date, end_date: date
) -> dict[str, Any]:
    """Get historical data for a stock index ticker from the API."""
    encoded_ticker = urllib.parse.quote(ticker)
    url = f"{BASE_URL}/historical-price-full/index/{encoded_ticker}"

    response = requests.get(
        url,
        params={
            "apikey": api_key,
            "from": start_date.strftime("%Y-%m-%d"),
            "to": end_date.strftime("%Y-%m-%d"),
        },
    )
    return response.json()


def get_forex_ticker(
    api_key: str, ticker: str, start_date: date, end_date: date
) -> dict[str, Any]:
    """Get historical data for a foreign exchange ticker from the API."""
    encoded_ticker = urllib.parse.quote(ticker)
    url = f"{BASE_URL}/historical-price-full/forex/{encoded_ticker}"

    response = requests.get(
        url,
        params={
            "apikey": api_key,
            "from": start_date.strftime("%Y-%m-%d"),
            "to": end_date.strftime("%Y-%m-%d"),
        },
    )
    return response.json()


def get_macro_data(api_key: str, start_date: date, end_date: date) -> list[dict]:
    """Pull macroeconomic data from start_date to end_date (inclusive)."""
    logging.info(f"Downloading market data from {start_date} to {end_date}")
    macro_data: list[dict[str, Any]] = []

    for ticker in INDEX_TICKERS:
        ticker_data = get_index_ticker(
            api_key=api_key,
            ticker=ticker,
            start_date=start_date,
            end_date=end_date,
        )
        # Exchanges are closed on weekends and some holidays, so these calls
        # could return nothing
        if ticker_data:
            macro_data.append(ticker_data)

    for ticker in FOREX_TICKERS:
        ticker_data = get_forex_ticker(
            api_key=api_key,
            ticker=ticker,
            start_date=start_date,
            end_date=end_date,
        )
        # Exchanges are closed on weekends and some holidays, so these calls
        # could return nothing
        if ticker_data:
            macro_data.append(ticker_data)

    logging.info("Market data successfully downloaded")
    return [
        {
            "symbol": row["symbol"],
            "market_date": day["date"],
            "open": day["open"],
            "close": day["close"],
            "adj_close": day["adjClose"],
            "high": day["high"],
            "low": day["low"],
            "volume": day["volume"],
        }
        for row in macro_data
        for day in row["historical"]
    ]


def load_data_to_bq(destination_table: str, macro_data: list[dict]):
    """Load downloaded data to BQ table."""
    credentials, project = google.auth.default(
        scopes=[
            "https://www.googleapis.com/auth/cloud-platform",
            "https://www.googleapis.com/auth/bigquery",
        ]
    )
    client = bigquery.Client(credentials=credentials, project=project)
    job_config = bigquery.LoadJobConfig(
        schema=[
            bigquery.SchemaField("symbol", "STRING"),
            bigquery.SchemaField("market_date", "DATE"),
            bigquery.SchemaField("open", "NUMERIC"),
            bigquery.SchemaField("close", "NUMERIC"),
            bigquery.SchemaField("adj_close", "NUMERIC"),
            bigquery.SchemaField("high", "NUMERIC"),
            bigquery.SchemaField("low", "NUMERIC"),
            bigquery.SchemaField("volume", "INTEGER"),
        ],
        autodetect=False,
        write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
        source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
    )
    job = client.load_table_from_json(
        macro_data, destination_table, job_config=job_config
    )
    job.result()


def main():
    """Either get and load a single day of data or backfill the entire table."""
    args = parser.parse_args()
    assert args.api_key, "You must provide an API key"

    if args.backfill:
        assert (
            args.start_date is not None and args.end_date is not None
        ), "You must provide a start and end date to backfill"
        assert args.market_date is None, "market date does not apply to backfills"
        start_date = datetime.strptime(args.start_date, "%Y-%m-%d").date()
        end_date = datetime.strptime(args.end_date, "%Y-%m-%d").date()
        macro_data = get_macro_data(args.api_key, start_date, end_date)
        destination_table = f"{args.destination_project}.{args.destination_dataset}.{args.destination_table}"

    else:
        assert (
            args.start_date is None and args.end_date is None
        ), "Start and end date only apply to backfills"
        assert (
            args.market_date is not None
        ), "You must provide a market date or --backfill + start and end date"
        market_date = end_date = datetime.strptime(args.market_date, "%Y-%m-%d").date()
        macro_data = get_macro_data(args.api_key, market_date, market_date)
        partition_str = market_date.strftime("%Y%m%d")
        destination_table = f"{args.destination_project}.{args.destination_dataset}.{args.destination_table}${partition_str}"

    # On a weekend or holiday we might not get any data at all
    if macro_data:
        load_data_to_bq(
            destination_table=destination_table,
            macro_data=macro_data,
        )
        logging.info(f"Loaded data to {destination_table}")
    else:
        logging.info("No data returned")


if __name__ == "__main__":
    main()



"""Methods for working with stable table schemas."""

import json
import tarfile
import urllib.request
from dataclasses import dataclass
from functools import cache
from io import BytesIO
from itertools import groupby
from typing import List

from bigquery_etl.config import ConfigLoader
from bigquery_etl.dryrun import DryRun


@dataclass
class SchemaFile:
    """Container for metadata about a JSON schema and corresponding BQ table."""

    schema: dict
    schema_id: str
    bq_dataset_family: str
    bq_table: str
    document_namespace: str
    document_type: str
    document_version: int

    @property
    def bq_table_unversioned(self):
        """Return table_id with version suffix stripped."""
        return "_".join(self.bq_table.split("_")[:-1])

    @property
    def stable_table(self):
        """Return BQ stable table name in <dataset>.<table> form."""
        return f"{self.bq_dataset_family}_stable.{self.bq_table}"

    @property
    def user_facing_view(self):
        """Return user-facing view name in <dataset>.<view> form."""
        return f"{self.bq_dataset_family}.{self.bq_table_unversioned}"

    @property
    def sortkey(self):
        """Return variant of stable_table with zero-padded version for sorting."""
        return (
            "_".join(self.stable_table.split("_")[:-1]) + f"{self.document_version:04d}"
        )


def prod_schemas_uri():
    """Return URI for the schemas tarball deployed to shared-prod.

    We construct a fake query and send it to the dry run service in order
    to read dataset labels, which contains the commit hash associated
    with the most recent production schemas deploy.
    """
    dryrun = DryRun(
        "moz-fx-data-shared-prod/telemetry_derived/foo/query.sql", content="SELECT 1"
    )
    build_id = dryrun.get_dataset_labels()["schemas_build_id"]
    commit_hash = build_id.split("_")[-1]
    mps_uri = ConfigLoader.get("schema", "mozilla_pipeline_schemas_uri")
    return f"{mps_uri}/archive/{commit_hash}.tar.gz"


@cache
def get_stable_table_schemas() -> List[SchemaFile]:
    """Fetch last schema metadata per doctype by version."""
    schemas_uri = prod_schemas_uri()
    with urllib.request.urlopen(schemas_uri) as f:
        tarbytes = BytesIO(f.read())

    schemas = []
    with tarfile.open(fileobj=tarbytes, mode="r:gz") as tar:
        for tarinfo in tar:
            if tarinfo.name.endswith(".schema.json"):
                *_, document_namespace, document_type, basename = tarinfo.name.split(
                    "/"
                )
                version = int(basename.split(".")[1])
                schema = json.load(tar.extractfile(tarinfo.name))  # type: ignore
                bq_schema = {}

                # Schemas without `bq_dataset_family` and `bq_table` metadata (like glean/glean)
                # do not have corresponding BQ tables, so we skip them here.
                pipeline_meta = schema.get("mozPipelineMetadata", None)
                if (
                    pipeline_meta is None
                    or "bq_dataset_family" not in pipeline_meta
                    or "bq_table" not in pipeline_meta
                ):
                    continue

                try:
                    bq_schema_file = tar.extractfile(
                        tarinfo.name.replace(".schema.json", ".bq")
                    )
                    bq_schema = json.load(bq_schema_file)  # type: ignore
                except KeyError as e:
                    print(f"Cannot get Bigquery schema for {tarinfo.name}: {e}")

                schemas.append(
                    SchemaFile(
                        schema=bq_schema,
                        schema_id=schema.get("$id", ""),
                        bq_dataset_family=pipeline_meta["bq_dataset_family"],
                        bq_table=pipeline_meta["bq_table"],
                        document_namespace=document_namespace,
                        document_type=document_type,
                        document_version=version,
                    )
                )

    # Exclude doctypes maintained in separate projects.
    for prefix in ConfigLoader.get("schema", "skip_prefixes", fallback=[]):
        schemas = [
            schema
            for schema in schemas
            if not schema.document_namespace.startswith(prefix)
        ]

    # Retain only the highest version per doctype.
    schemas = sorted(
        schemas,
        key=lambda t: f"{t.document_namespace}/{t.document_type}/{t.document_version:03d}",
    )
    schemas = [
        last
        for k, (*_, last) in groupby(
            schemas, lambda t: f"{t.document_namespace}/{t.document_type}"
        )
    ]

    return schemas



"""Feature usage table generation."""

import os
import shutil
from pathlib import Path

import click
import yaml
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.schema import SCHEMA_FILE, Schema
from bigquery_etl.util.common import write_sql

FILE_PATH = Path(os.path.dirname(__file__))
BASE_DIR = Path(FILE_PATH).parent.parent
TEMPLATE_CONFIG = FILE_PATH / "templating.yaml"


def generate_query(project, dataset, destination_table, write_dir):
    """Generate feature usage table query."""
    with open(TEMPLATE_CONFIG, "r") as f:
        render_kwargs = yaml.safe_load(f) or {}
    env = Environment(loader=FileSystemLoader(str(FILE_PATH / "templates")))
    template = env.get_template("query.sql")

    write_sql(
        write_dir / project,
        f"{project}.{dataset}.{destination_table}",
        "query.sql",
        reformat(template.render(**render_kwargs)),
    )


def generate_view(project, dataset, destination_table, write_dir):
    """Generate feature usage table view."""
    view_name = destination_table.split("_v")[0]
    view_dataset = dataset.split("_derived")[0]

    sql = reformat(
        f"""
        CREATE OR REPLACE VIEW `{project}.{view_dataset}.{view_name}` AS
        SELECT
            *
        FROM
            `{project}.{dataset}.{destination_table}`
    """
    )

    write_sql(
        write_dir / project, f"{project}.{view_dataset}.{view_name}", "view.sql", sql
    )


def generate_metadata(project, dataset, destination_table, write_dir):
    """Copy metadata.yaml file to destination directory."""
    shutil.copyfile(
        FILE_PATH / "templates" / "metadata.yaml",
        write_dir / project / dataset / destination_table / "metadata.yaml",
    )


def generate_schema(project, dataset, destination_table, write_dir):
    """Generate the table schema."""
    # get schema
    table_schema = Schema.for_table(project, dataset, destination_table)
    query_schema = Schema.from_query_file(
        write_dir / project / dataset / destination_table / "query.sql",
        use_cloud_function=True,
    )
    query_schema.merge(table_schema)
    schema_path = write_dir / project / dataset / destination_table / SCHEMA_FILE
    query_schema.to_yaml_file(schema_path)


@click.command("generate")
@click.option(
    "--target-project",
    "--target_project",
    help="Which project the queries should be written to.",
    default="moz-fx-data-shared-prod",
)
@click.option(
    "--dataset",
    help="Which dataset the queries should be written to.",
    default="telemetry_derived",
)
@click.option(
    "--destination_table",
    "--destination-table",
    help="Name of the destination table.",
    default="feature_usage_v2",
)
@click.option(
    "--output-dir",
    "--output_dir",
    help="The location to write to. Defaults to sql/.",
    default=Path("sql"),
    type=click.Path(file_okay=False),
)
@use_cloud_function_option
@click.pass_context
def generate(
    ctx, target_project, dataset, destination_table, output_dir, use_cloud_function
):
    """Generate the feature usage table."""
    output_dir = Path(output_dir)
    generate_query(target_project, dataset, destination_table, output_dir)
    generate_view(target_project, dataset, destination_table, output_dir)
    generate_metadata(target_project, dataset, destination_table, output_dir)
    generate_schema(target_project, dataset, destination_table, output_dir)



"""Validate backfill entries."""

from pathlib import Path
from typing import List

from ..backfill.parse import DEFAULT_REASON, DEFAULT_WATCHER, Backfill, BackfillStatus


def validate_duplicate_entry_dates(
    backfill_entry: Backfill, backfills: list[Backfill]
) -> None:
    """Check if backfill entries have the same entry dates."""
    for b in backfills:
        if backfill_entry.entry_date == b.entry_date:
            raise ValueError(
                f"Duplicate backfill with entry date: {backfill_entry.entry_date}."
            )


def validate_excluded_dates(entry: Backfill) -> None:
    """Check if backfill excluded dates are sorted and have no duplicates."""
    if not entry.excluded_dates == sorted(entry.excluded_dates):
        raise ValueError(
            f"Existing backfill entry with excluded dates not sorted: {entry.entry_date}."
        )
    if not len(entry.excluded_dates) == len(set(entry.excluded_dates)):
        raise ValueError(
            f"Existing backfill entry with duplicate excluded dates: {entry.entry_date}."
        )


def validate_default_reason(entry: Backfill) -> None:
    """Check if backfill reason is the same as default."""
    if entry.reason == DEFAULT_REASON:
        raise ValueError(f"Default reason found: {entry.reason}.")


def validate_default_watchers(entry: Backfill) -> None:
    """Check if backfill watcher is the same as default."""
    if DEFAULT_WATCHER in entry.watchers:
        raise ValueError(f"Default watcher found: ({entry.watchers}).")


def validate_entries_are_sorted(backfills: List[Backfill]) -> None:
    """Check if list of backfill entries are sorted by entry dates."""
    entry_dates = [b.entry_date for b in backfills]
    if not entry_dates == sorted(entry_dates, reverse=True):
        raise ValueError("Backfill entries are not sorted by entry dates")


def validate_file(file: Path) -> None:
    """Validate all entries from a given backfill.yaml file."""
    backfills = Backfill.entries_from_file(file)
    validate_entries(backfills)


def validate_duplicate_entry_with_initiate_status(
    backfill_entry: Backfill, backfills: list
) -> None:
    """Check if list of backfill entries have more than one entry with Initiate Status."""
    if backfill_entry.status == BackfillStatus.INITIATE:
        for b in backfills:
            if b.status == BackfillStatus.INITIATE:
                raise ValueError(
                    "Backfill entries cannot contain more than one entry with Initiate status"
                )


def validate_entries(backfills: list) -> None:
    """Validate a list of backfill entries."""
    for i, backfill_entry in enumerate(backfills):
        validate_default_watchers(backfill_entry)
        validate_default_reason(backfill_entry)
        validate_duplicate_entry_dates(backfill_entry, backfills[i + 1 :])
        validate_excluded_dates(backfill_entry)
        validate_duplicate_entry_with_initiate_status(
            backfill_entry, backfills[i + 1 :]
        )
    validate_entries_are_sorted(backfills)
